from collections import UserList
from collections.abc import Callable
from datetime import datetime, timezone
from typing import Any, cast

from cognite.neat._client.client import NeatClient
from cognite.neat._client.data_classes import SpaceStatisticsResponse
from cognite.neat._config import NeatConfig
from cognite.neat._data_model._shared import OnSuccess, OnSuccessIssuesChecker, OnSuccessResultProducer
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.deployer.data_classes import DeploymentResult
from cognite.neat._data_model.deployer.deployer import SchemaDeployer
from cognite.neat._data_model.exporters import DMSExporter, DMSFileExporter
from cognite.neat._data_model.exporters._api_exporter import DMSAPIExporter
from cognite.neat._data_model.exporters._table_exporter.exporter import DMSTableExporter
from cognite.neat._data_model.importers import DMSImporter, DMSTableImporter
from cognite.neat._data_model.models.dms import RequestSchema as PhysicalDataModel
from cognite.neat._data_model.models.dms._limits import SchemaLimits
from cognite.neat._exceptions import DataModelCreateException, DataModelImportException
from cognite.neat._issues import IssueList
from cognite.neat._state_machine._states import EmptyState, PhysicalState, State
from cognite.neat._utils.text import NEWLINE

from ._provenance import Change, Provenance

Agents = DMSExporter | DMSTableImporter | DMSImporter


class NeatStore:
    def __init__(self, config: NeatConfig, client: NeatClient) -> None:
        self.physical_data_model = DataModelList()
        self.provenance = Provenance()
        self.state: State = EmptyState()
        self._client = client
        self._config = config

        # Placeholder for CDF schema and limit snapshot
        self._cdf_snapshot: SchemaSnapshot | None = None
        self._cdf_limits: SchemaLimits | None = None
        self._cdf_space_statistics: SpaceStatisticsResponse | None = None

    @property
    def cdf_space_statistics(self) -> SpaceStatisticsResponse:
        if self._cdf_space_statistics is None:
            self._cdf_space_statistics = self._client.statistics.space_statistics(
                [space.space for space in self.cdf_snapshot.spaces.keys()]
            )
        return self._cdf_space_statistics

    @property
    def cdf_limits(self) -> SchemaLimits:
        if not self._cdf_limits:
            self._cdf_limits = SchemaLimits.from_api_response(self._client.statistics.project())
        return self._cdf_limits

    @property
    def cdf_snapshot(self) -> SchemaSnapshot:
        if not self._cdf_snapshot:
            self._cdf_snapshot = SchemaSnapshot.fetch_entire_cdf(self._client)
        return self._cdf_snapshot

    def read_physical(self, reader: DMSImporter, on_success: OnSuccess | None = None) -> None:
        """Read object from the store"""
        self._can_agent_do_activity(reader)

        change, data_model = self._do_activity(reader.to_data_model, on_success)

        if data_model:
            change.target_entity = self.physical_data_model.generate_reference(cast(PhysicalDataModel, data_model))
            self.physical_data_model.append(data_model)
            self.state = self.state.transition(reader)
            change.target_state = self.state

        self.provenance.append(change)

    def write_physical(self, writer: DMSExporter, on_success: OnSuccess | None = None, **kwargs: Any) -> None:
        """Write object into the store"""
        self._can_agent_do_activity(writer)

        activity: Callable
        if isinstance(writer, DMSFileExporter):
            activity = writer.export_to_file
            if not kwargs.get("file_path"):
                raise RuntimeError("file_path must be provided when using a DMSFileExporter")
        else:
            activity = writer.export

        data_model = self._gather_data_model(writer)

        change, _ = self._do_activity(activity, on_success, data_model=data_model, **kwargs)

        if not change.issues:
            change.target_entity = "ExternalEntity"
            self.state = self.state.transition(writer)
            change.target_state = self.state

        self.provenance.append(change)

        if (
            isinstance(writer, DMSAPIExporter)
            and isinstance(on_success, SchemaDeployer)
            and not on_success.options.dry_run
        ):
            # Update CDF snapshot and space statistics after deployment
            self._cdf_snapshot = SchemaSnapshot.fetch_entire_cdf(self._client)
            self._cdf_space_statistics = self._client.statistics.space_statistics(
                [space.space for space in self.cdf_snapshot.spaces.keys()]
            )

    def cdf_analyze(self, on_success: OnSuccess) -> None:
        """Analyze the entity of CDF data models"""
        change, _ = self._do_activity(lambda: self.cdf_snapshot, on_success)
        self.provenance.append(change)

    def _gather_data_model(self, writer: DMSExporter) -> PhysicalDataModel:
        """Gather the current physical data model from the store

        Args:
            writer (DMSExporter): The exporter that will be used to write the data model.
        """
        # getting provenance of the last successful physical data model read
        change = self.provenance.last_physical_data_model_read()

        if not change:
            raise RuntimeError("No successful physical data model read found in provenance.")

        # We do not want to modify the data model for API representations
        if not isinstance(writer, DMSTableExporter):
            return self.physical_data_model[-1]

        # This will handle data model that are partially and require to be converted to
        # tabular representation to include all containers referenced by views.
        copy = self.physical_data_model[-1].model_copy(deep=True)
        container_refs = {container.as_reference() for container in copy.containers}

        for view in copy.views:
            for container in view.used_containers:
                if container not in container_refs and (cdf_container := self.cdf_snapshot.containers.get(container)):
                    copy.containers.append(cdf_container)
                    container_refs.add(container)

        return copy

    def _can_agent_do_activity(self, agent: Agents) -> None:
        """Validate if activity can be performed in the current state and considering provenance"""
        if not self.state.can_transition(agent):
            # specific error messages for common mistakes
            if isinstance(agent, DMSImporter) and isinstance(self.state, PhysicalState):
                raise RuntimeError(
                    "⚠️ Cannot read data model, there is already a data model in the session!"
                    f"{NEWLINE}Start a new session to read a new data model."
                )

            if isinstance(agent, DMSExporter) and isinstance(self.state, EmptyState):
                raise RuntimeError(
                    "⚠️ Cannot write data model, there is no data model in the session!"
                    f"{NEWLINE}Read a data model first!"
                )
            raise RuntimeError(f"Cannot run {type(agent).__name__} in state {self.state}")

        if (
            isinstance(agent, DMSExporter)
            and self.provenance.last_change
            and (error_count := self.provenance.last_change.error_count) > 0
        ):
            raise RuntimeError(
                f"⚠️ Cannot write data model, the model has {error_count} errors!"
                f"{NEWLINE}Resolve issues before exporting the data model."
                f"{NEWLINE}You can inspect issues using neat.issues"
            )

        # need implementation of checking if required predecessor activities have been done
        # this will be done by running self.provenance.can_agent_do_activity(agent)

    def _do_activity(
        self, activity: Callable, on_success: OnSuccess | None = None, **kwargs: Any
    ) -> tuple[Change, PhysicalDataModel | None]:
        """Execute activity and capture timing, results, and issues"""
        start = datetime.now(timezone.utc)
        created_data_model: PhysicalDataModel | None = None
        issues = IssueList()
        errors = IssueList()
        deployment_result: DeploymentResult | None = None

        try:
            created_data_model = activity(**kwargs)
            if created_data_model and on_success:
                on_success.run(created_data_model)
                if isinstance(on_success, OnSuccessIssuesChecker):
                    issues.extend(on_success.issues)
                elif isinstance(on_success, OnSuccessResultProducer):
                    deployment_result = on_success.result
                else:
                    raise RuntimeError(f"Unknown OnSuccess type {type(on_success).__name__}")

        # we catch import exceptions to capture issues and errors in provenance
        except DataModelImportException as e:
            errors.extend(e.errors)

        except DataModelCreateException as e:
            errors.extend(e.errors)

        # these are all other errors, such as missing file, wrong format, etc.
        except Exception as e:
            raise e

        end = datetime.now(timezone.utc)

        return Change(
            start=start,
            end=end,
            source_state=self.state,
            agent=type(activity.__self__).__name__ if hasattr(activity, "__self__") else "UnknownAgent",
            issues=issues,
            errors=errors,
            result=deployment_result,
            activity=Change.standardize_activity_name(activity.__name__, start, end),
        ), created_data_model


class DataModelList(UserList[PhysicalDataModel]):
    def iteration(self, data_model: PhysicalDataModel) -> int:
        """Get iteration number for data model"""
        for i, existing in enumerate(self):
            if existing.data_model == data_model.data_model:
                return i + 2
        return 1

    def generate_reference(self, data_model: PhysicalDataModel) -> str:
        """Generate reference string for data model based on iteration"""
        space = data_model.data_model.space
        external_id = data_model.data_model.external_id
        version = data_model.data_model.version
        iteration = self.iteration(data_model)

        return f"physical/{space}/{external_id}/{version}/{iteration}"

    def get_by_reference(self, reference: str) -> PhysicalDataModel | None:
        """Get data model by reference string"""

        raise NotImplementedError("Not implemented yet")



from datetime import date, datetime

from pydantic import Field

from cognite.neat._data_model._constants import XML_SCHEMA_NAMESPACE
from cognite.neat._data_model._identifiers import URI

from ._base import Entity


class UnitEntity(Entity): ...


class EnumCollectionEntity(Entity): ...


class DataType(Entity):
    prefix: str = Field("xsd", frozen=True)

    @property
    def python(self) -> type:
        raise NotImplementedError()

    @property
    def xsd(self) -> URI:
        return XML_SCHEMA_NAMESPACE[self.suffix]


class Boolean(DataType):
    suffix: str = Field("boolean", frozen=True)

    @property
    def python(self) -> type:
        return bool


class Float(DataType):
    suffix: str = Field("float", frozen=True)
    unit: UnitEntity | None = None

    @property
    def python(self) -> type:
        return float


class Double(DataType):
    suffix: str = Field("double", frozen=True)
    unit: UnitEntity | None = None

    @property
    def python(self) -> type:
        return float


class Integer(DataType):
    suffix: str = Field("integer", frozen=True)
    unit: UnitEntity | None = None

    @property
    def python(self) -> type:
        return int


class Long(DataType):
    suffix: str = Field("long", frozen=True)
    unit: UnitEntity | None = None

    @property
    def python(self) -> type:
        return int


class String(DataType):
    suffix: str = Field("string", frozen=True)
    max_text_size: int | None = Field(
        None,
        alias="maxTextSize",
        description="Specifies the maximum size in bytes of the text property, when encoded with utf-8.",
    )

    @property
    def python(self) -> type:
        return str


class AnyURI(DataType):
    suffix: str = Field("anyURI", frozen=True)

    @property
    def python(self) -> type:
        return str


class Date(DataType):
    suffix: str = Field("date", frozen=True)

    @property
    def python(self) -> type:
        return date


class DateTime(DataType):
    suffix: str = Field("dateTime", frozen=True)

    @property
    def python(self) -> type:
        return datetime


class DateTimeStamp(DataType):
    suffix: str = Field("dateTimeStamp", frozen=True)

    @property
    def python(self) -> type:
        return datetime


# CDF Specific extensions of XSD types


class Timeseries(DataType):
    suffix: str = Field("timeseries", frozen=True)


class File(DataType):
    suffix: str = Field("string", frozen=True)


class Sequence(DataType):
    suffix: str = Field("sequence", frozen=True)


class Json(DataType):
    suffix: str = Field("json", frozen=True)

    @property
    def python(self) -> type:
        return dict


class Enum(DataType):
    suffix: str = Field("enum", frozen=True)
    collection: EnumCollectionEntity
    unknown_value: str | None = Field(None, alias="unknownValue")



import json
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Literal, TypeVar, cast, overload

from pydantic import BaseModel, TypeAdapter, ValidationError

from cognite.neat._data_model.models.dms import (
    Constraint,
    ConstraintAdapter,
    ContainerPropertyDefinition,
    ContainerReference,
    ContainerRequest,
    DataModelRequest,
    Index,
    IndexAdapter,
    NodeReference,
    RequestSchema,
    SpaceRequest,
    UniquenessConstraintDefinition,
    ViewReference,
    ViewRequest,
    ViewRequestProperty,
    ViewRequestPropertyAdapter,
)
from cognite.neat._data_model.models.dms._constants import DATA_MODEL_DESCRIPTION_MAX_LENGTH
from cognite.neat._data_model.models.entities import ParsedEntity, parse_entity
from cognite.neat._exceptions import DataModelImportException
from cognite.neat._issues import ModelSyntaxError
from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.validation import ValidationContext, humanize_validation_error

from .data_classes import (
    CREATOR_KEY,
    CREATOR_MARKER,
    DMSContainer,
    DMSEnum,
    DMSNode,
    DMSProperty,
    DMSView,
    EntityTableFilter,
    RAWFilterTableFilter,
    TableDMS,
    TableViewFilter,
)
from .source import TableSource

T_BaseModel = TypeVar("T_BaseModel", bound=BaseModel)


@dataclass
class ReadViewProperty:
    prop_id: str
    row_no: int
    view_property: ViewRequestProperty


@dataclass
class ReadContainerProperty:
    prop_id: str
    row_no: int
    container_property: ContainerPropertyDefinition


@dataclass
class ReadIndex:
    prop_id: str
    order: int | None
    row_no: int
    index_id: str
    index: Index


@dataclass
class ReadConstraint:
    prop_id: str
    order: int | None
    row_no: int
    constraint_id: str
    constraint: Constraint


@dataclass
class ReadProperties:
    """Read properties from the properties table.

    Attributes:
        container: A mapping from container entity to a mapping of property identifier to container property definition.
        view: A mapping from view entity to a mapping of property identifier to view property definition.
        indices: A mapping from (container entity, index identifier) to a list of read indices
        constraints: A mapping from (container entity, constraint identifier) to a list of read constraints
    """

    container: dict[tuple[ParsedEntity, str], list[ReadContainerProperty]] = field(
        default_factory=lambda: defaultdict(list)
    )
    view: dict[tuple[ParsedEntity, str], list[ReadViewProperty]] = field(default_factory=lambda: defaultdict(list))
    indices: dict[tuple[ParsedEntity, str], list[ReadIndex]] = field(default_factory=lambda: defaultdict(list))
    constraints: dict[tuple[ParsedEntity, str], list[ReadConstraint]] = field(default_factory=lambda: defaultdict(list))


@dataclass
class ProcessedProperties:
    container: dict[ParsedEntity, dict[str, ContainerPropertyDefinition]] = field(
        default_factory=lambda: defaultdict(dict)
    )
    view: dict[ParsedEntity, dict[str, ViewRequestProperty]] = field(default_factory=lambda: defaultdict(dict))
    indices: dict[ParsedEntity, dict[str, Index]] = field(default_factory=lambda: defaultdict(dict))
    constraints: dict[ParsedEntity, dict[str, Constraint]] = field(default_factory=lambda: defaultdict(dict))


class DMSTableReader:
    """Reads a TableDMS object and converts it to a RequestSchema.


    Args:
        default_space (str): The default space to use when no space is given in an entity.
        default_version (str): The default version to use when no version is given in an entity.
        source (TableSource): The source of the table data, used for error reporting.

    Raises:
        DataModelImportError: If there are any errors in the data model.

    Attributes:
        errors (list[ModelSyntaxError]): A list of errors encountered during parsing.

    Class Attributes:
        Sheets: This is used to create error messages. It ensures that the column names matches
            the names in the table, even if they are renamed in the code.
        PropertyColumns: This is used to create error messages for the properties table.
            It ensures that the column names matches the names in the table, even if they are renamed in the code.
        ContainerColumns: This is used to create error messages for the containers table.
            It ensures that the column names matches the names in the table, even if they are renamed in the code.
        ViewColumns: This is used to create error messages for the views table.
            It ensures that the column names matches the names in the table, even if they are renamed in the code.

    """

    CELL_MISSING = "#N/A"

    # The following classes are used when creating error messages. They ensure that the column names
    # matches the names in the table, even if they are renamed in the code.
    # Note that this is not a complete list of all columns, only those that are used in error messages.
    class Sheets:
        metadata = cast(str, TableDMS.model_fields["metadata"].validation_alias)
        properties = cast(str, TableDMS.model_fields["properties"].validation_alias)
        containers = cast(str, TableDMS.model_fields["containers"].validation_alias)
        views = cast(str, TableDMS.model_fields["views"].validation_alias)
        nodes = cast(str, TableDMS.model_fields["nodes"].validation_alias)

    class PropertyColumns:
        view = cast(str, DMSProperty.model_fields["view"].validation_alias)
        view_property = cast(str, DMSProperty.model_fields["view_property"].validation_alias)
        connection = cast(str, DMSProperty.model_fields["connection"].validation_alias)
        value_type = cast(str, DMSProperty.model_fields["value_type"].validation_alias)
        min_count = cast(str, DMSProperty.model_fields["min_count"].validation_alias)
        max_count = cast(str, DMSProperty.model_fields["max_count"].validation_alias)
        default = cast(str, DMSProperty.model_fields["default"].validation_alias)
        auto_increment = cast(str, DMSProperty.model_fields["auto_increment"].validation_alias)
        container = cast(str, DMSProperty.model_fields["container"].validation_alias)
        container_property = cast(str, DMSProperty.model_fields["container_property"].validation_alias)
        container_property_name = cast(str, DMSProperty.model_fields["container_property_name"].validation_alias)
        container_property_description = cast(
            str, DMSProperty.model_fields["container_property_description"].validation_alias
        )
        index = cast(str, DMSProperty.model_fields["index"].validation_alias)
        constraint = cast(str, DMSProperty.model_fields["constraint"].validation_alias)

    class ContainerColumns:
        container = cast(str, DMSContainer.model_fields["container"].validation_alias)
        constraint = cast(str, DMSContainer.model_fields["constraint"].validation_alias)

    class ViewColumns:
        view = cast(str, DMSView.model_fields["view"].validation_alias)
        filter = cast(str, DMSView.model_fields["filter"].validation_alias)

    def __init__(self, default_space: str, default_version: str, source: TableSource) -> None:
        self.default_space = default_space
        self.default_version = default_version
        self.source = source
        self.errors: list[ModelSyntaxError] = []

    def read_tables(self, tables: TableDMS) -> RequestSchema:
        space_request = self.read_space(self.default_space)
        node_types = self.read_nodes(tables.nodes)
        enum_collections = self.read_enum_collections(tables.enum)
        container_ref_by_entity = self.read_entity_by_container_ref(tables.containers)
        view_ref_by_entity = self.read_entity_by_view_ref(tables.views)
        read = self.read_properties(tables.properties, enum_collections, container_ref_by_entity, view_ref_by_entity)
        processed = self.process_properties(read)
        containers = self.read_containers(tables.containers, processed)
        views, valid_view_entities = self.read_views(tables.views, processed.view)
        data_model = self.read_data_model(tables, valid_view_entities)

        if self.errors:
            raise DataModelImportException(self.errors) from None
        return RequestSchema(
            dataModel=data_model, views=views, containers=containers, spaces=[space_request], nodeTypes=node_types
        )

    def read_space(self, space: str) -> SpaceRequest:
        space_request = self._validate_obj(SpaceRequest, {"space": space}, (self.Sheets.metadata,), field_name="value")
        if space_request is None:
            # If space is invalid, we stop parsing to avoid raising an error for every place the space is used.
            raise DataModelImportException(self.errors) from None
        return space_request

    def read_nodes(self, nodes: list[DMSNode]) -> list[NodeReference]:
        node_refs: list[NodeReference] = []
        for row_no, row in enumerate(nodes):
            data = self._create_node_ref(row.node)
            instantiated = self._validate_obj(NodeReference, data, (self.Sheets.nodes, row_no))
            if instantiated is not None:
                node_refs.append(instantiated)
        return node_refs

    @staticmethod
    def read_enum_collections(enum_rows: list[DMSEnum]) -> dict[str, dict[str, Any]]:
        enum_collections: dict[str, dict[str, Any]] = defaultdict(dict)
        for row in enum_rows:
            enum_collections[row.collection][row.value] = {
                "name": row.name,
                "description": row.description,
            }
        return enum_collections

    def read_entity_by_container_ref(self, containers: list[DMSContainer]) -> dict[ContainerReference, ParsedEntity]:
        entity_by_container_ref: dict[ContainerReference, ParsedEntity] = {}
        for container in containers:
            data = self._create_container_ref(container.container)
            try:
                container_ref = ContainerReference.model_validate(data)
            except ValidationError:
                # Error will be reported when reading the containers
                continue
            entity_by_container_ref[container_ref] = container.container
        return entity_by_container_ref

    def read_entity_by_view_ref(self, views: list[DMSView]) -> dict[ViewReference, ParsedEntity]:
        entity_by_view_ref: dict[ViewReference, ParsedEntity] = {}
        for view in views:
            data = self._create_view_ref(view.view)
            try:
                view_ref = ViewReference.model_validate(data)
            except ValidationError:
                # Error will be reported when reading the views
                continue
            entity_by_view_ref[view_ref] = view.view
        return entity_by_view_ref

    def read_properties(
        self,
        properties: list[DMSProperty],
        enum_collections: dict[str, dict[str, Any]],
        container_ref_by_entity: dict[ContainerReference, ParsedEntity],
        view_ref_by_entity: dict[ViewReference, ParsedEntity],
    ) -> ReadProperties:
        read = ReadProperties()
        view_entities = set(view_ref_by_entity.values())
        container_entities = set(container_ref_by_entity.values())
        for row_no, prop in enumerate(properties):
            self._process_view_property(prop, read, row_no, view_ref_by_entity, view_entities)
            if prop.container is None or prop.container_property is None:
                # This is when the property is an edge or reverse direct relation property.
                continue
            self._process_container_property(
                prop, read, enum_collections, row_no, container_ref_by_entity, container_entities
            )
            self._process_index(prop, read, row_no, container_ref_by_entity, container_entities)
            self._process_constraint(prop, read, row_no, container_ref_by_entity, container_entities)
        return read

    def process_properties(self, read: ReadProperties) -> ProcessedProperties:
        return ProcessedProperties(
            container=self.create_container_properties(read),
            view=self.create_view_properties(read),
            indices=self.create_indices(read),
            constraints=self.create_constraints(read),
        )

    def create_container_properties(
        self, read: ReadProperties
    ) -> dict[ParsedEntity, dict[str, ContainerPropertyDefinition]]:
        container_props: dict[ParsedEntity, dict[str, ContainerPropertyDefinition]] = defaultdict(dict)
        for (container_entity, prop_id), prop_list in read.container.items():
            if len(prop_list) == 0:
                # Should not happen
                continue
            container_props[container_entity][prop_id] = prop_list[0].container_property
            if len(prop_list) > 1 and self._are_definitions_different(prop_list):
                # If multiple view properties are mapping to the same container property,
                # the container property definitions must be the same.
                rows_str = humanize_collection(
                    [self.source.adjust_row_number(self.Sheets.properties, p.row_no) for p in prop_list]
                )
                container_columns_str = humanize_collection(
                    [
                        self.PropertyColumns.connection,
                        self.PropertyColumns.value_type,
                        self.PropertyColumns.min_count,
                        self.PropertyColumns.max_count,
                        self.PropertyColumns.default,
                        self.PropertyColumns.auto_increment,
                        self.PropertyColumns.container_property_name,
                        self.PropertyColumns.container_property_description,
                        self.PropertyColumns.index,
                        self.PropertyColumns.constraint,
                    ]
                )
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location((self.Sheets.properties,))} "
                            f"when the column {self.PropertyColumns.container!r} and "
                            f"{self.PropertyColumns.container_property!r} are the same, "
                            f"all the container columns ({container_columns_str}) must be the same. "
                            f"Inconsistent definitions for container '{container_entity!s} "
                            f"and {prop_id!r}' found in rows {rows_str}."
                        )
                    )
                )
        return container_props

    def _are_definitions_different(self, prop_list: list[ReadContainerProperty]) -> bool:
        if len(prop_list) < 2:
            return False
        first_def = prop_list[0].container_property
        for prop in prop_list[1:]:
            if first_def != prop.container_property:
                return True
        return False

    def create_view_properties(self, read: ReadProperties) -> dict[ParsedEntity, dict[str, ViewRequestProperty]]:
        view_props: dict[ParsedEntity, dict[str, ViewRequestProperty]] = defaultdict(dict)
        for (view_entity, prop_id), prop_list in read.view.items():
            if len(prop_list) == 0:
                # Should not happen
                continue
            view_props[view_entity][prop_id] = prop_list[0].view_property
            if len(prop_list) > 1:
                # Safeguard against duplicated rows for view properties.
                rows_str = humanize_collection(
                    [self.source.adjust_row_number(self.Sheets.properties, p.row_no) for p in prop_list]
                )
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location((self.Sheets.properties,))} the combination of columns "
                            f"{self.PropertyColumns.view!r} and {self.PropertyColumns.view_property!r} must be unique. "
                            f"Duplicated entries for view '{view_entity!s}' and "
                            f"property '{prop_id!s}' found in rows {rows_str}."
                        )
                    )
                )

        return view_props

    def create_indices(self, read: ReadProperties) -> dict[ParsedEntity, dict[str, Index]]:
        indices: dict[ParsedEntity, dict[str, Index]] = defaultdict(dict)
        for (container_entity, index_id), index_list in read.indices.items():
            if len(index_list) == 0:
                continue
            # Remove duplicates based on prop_id, keeping the first occurrence
            # Note that we have already validated that the index definitions are the same
            index_list = list({read_index.prop_id: read_index for read_index in index_list}.values())
            index = index_list[0].index
            if len(index_list) == 1:
                indices[container_entity][index_id] = index
                continue
            if missing_order := [idx for idx in index_list if idx.order is None]:
                row_str = humanize_collection(
                    [self.source.adjust_row_number(self.Sheets.properties, idx.row_no) for idx in missing_order]
                )
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In table {self.Sheets.properties!r} column {self.PropertyColumns.index!r}: "
                            f"the index {index_id!r} on container {container_entity!s} is defined on multiple "
                            f"properties. This requires the 'order' attribute to be set. It is missing in rows "
                            f"{row_str}."
                        )
                    )
                )
                continue
            index.properties = [idx.prop_id for idx in sorted(index_list, key=lambda x: x.order or 999)]
            indices[container_entity][index_id] = index
        return indices

    def create_constraints(self, read: ReadProperties) -> dict[ParsedEntity, dict[str, Constraint]]:
        constraints: dict[ParsedEntity, dict[str, Constraint]] = defaultdict(dict)
        for (container_entity, constraint_id), constraint_list in read.constraints.items():
            if len(constraint_list) == 0:
                continue
            # Remove duplicates based on prop_id, keeping the first occurrence
            # Note that we have already validated that the constraint definitions are the same
            constraint_list = list(
                {read_constraint.prop_id: read_constraint for read_constraint in constraint_list}.values()
            )
            constraint = constraint_list[0].constraint
            if len(constraint_list) == 1 or not isinstance(constraint, UniquenessConstraintDefinition):
                constraints[container_entity][constraint_id] = constraint
                continue
            if missing_order := [c for c in constraint_list if c.order is None]:
                row_str = humanize_collection(
                    [self.source.adjust_row_number(self.Sheets.properties, c.row_no) for c in missing_order]
                )
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In table {self.Sheets.properties!r} column {self.PropertyColumns.constraint!r}: "
                            f"the uniqueness constraint {constraint_id!r} on container {container_entity!s} is defined "
                            f"on multiple properties. This requires the 'order' attribute to be set. It is missing in "
                            f"rows {row_str}."
                        )
                    )
                )
                continue
            constraint.properties = [c.prop_id for c in sorted(constraint_list, key=lambda x: x.order or 999)]
            constraints[container_entity][constraint_id] = constraint
        return constraints

    def _process_view_property(
        self,
        prop: DMSProperty,
        read: ReadProperties,
        row_no: int,
        view_ref_by_entity: dict[ViewReference, ParsedEntity],
        view_entities: set[ParsedEntity],
    ) -> None:
        loc = (self.Sheets.properties, row_no)
        data = self.read_view_property(prop, loc)
        view_prop = self._validate_adapter(ViewRequestPropertyAdapter, data, loc)
        if view_prop is None:
            return None
        if prop.view in view_entities:
            read.view[(prop.view, prop.view_property)].append(ReadViewProperty(prop.view_property, row_no, view_prop))
            return None
        # The view entity was not found in the views table. This could either be because the view is missing,
        # or because either the view entity in the Properties table and the View table are specified with/without
        # default space/version inconsistently.
        try:
            view_ref = ViewReference.model_validate(self._create_view_ref(prop.view))
        except ValidationError:
            # Error will be reported when reading the views
            return None
        if view_ref in view_ref_by_entity:
            view_ref_entity = view_ref_by_entity[view_ref]
            read.view[(view_ref_entity, prop.view_property)].append(
                ReadViewProperty(prop.view_property, row_no, view_prop)
            )
        else:
            self.errors.append(
                ModelSyntaxError(
                    message=(
                        f"In {self.source.location(loc)} the View '{prop.view!s}' "
                        f"was not found in the {self.Sheets.views!r} table."
                    )
                )
            )
        return None

    def _process_container_property(
        self,
        prop: DMSProperty,
        read: ReadProperties,
        enum_collections: dict[str, dict[str, Any]],
        row_no: int,
        container_ref_by_entity: dict[ContainerReference, ParsedEntity],
        container_entities: set[ParsedEntity],
    ) -> None:
        loc = (self.Sheets.properties, row_no)
        data = self.read_container_property(prop, enum_collections, loc=loc)
        container_prop = self._validate_obj(ContainerPropertyDefinition, data, loc)
        if container_prop is None:
            return None
        if not (prop.container and prop.container_property):
            return None
        if prop.container in container_entities:
            read.container[(prop.container, prop.container_property)].append(
                ReadContainerProperty(prop.container_property, row_no, container_prop)
            )
            return None
        # The container entity was not found in the containers table. This could either be because the container
        # is missing, or because either the container entity in the Properties table and the Container table are
        # specified with/without default space/version inconsistently.
        try:
            container_ref = ContainerReference.model_validate(self._create_container_ref(prop.container))
        except ValidationError:
            # Error will be reported when reading the containers
            return None
        if container_ref in container_ref_by_entity:
            container_ref_entity = container_ref_by_entity[container_ref]
            read.container[(container_ref_entity, prop.container_property)].append(
                ReadContainerProperty(prop.container_property, row_no, container_prop)
            )
        # Container can be in CDF, this will be reported by a validator later.
        return None

    def _process_index(
        self,
        prop: DMSProperty,
        read: ReadProperties,
        row_no: int,
        container_ref_by_entity: dict[ContainerReference, ParsedEntity],
        container_entities: set[ParsedEntity],
    ) -> None:
        if prop.index is None or prop.container_property is None or prop.container is None:
            return

        loc = (self.Sheets.properties, row_no, self.PropertyColumns.index)
        for index in prop.index:
            data = self.read_index(index, prop.container_property)
            created = self._validate_adapter(IndexAdapter, data, loc)
            if created is None:
                continue
            order = self._read_order(index.properties, loc)

            if prop.container in container_entities:
                read.indices[(prop.container, index.suffix)].append(
                    ReadIndex(
                        prop_id=prop.container_property,
                        order=order,
                        row_no=row_no,
                        index_id=index.suffix,
                        index=created,
                    )
                )
                continue
            # The container entity was not found in the containers table. This could either be because the
            # container is missing, or because either the container entity in the Properties table and the
            # Container table are specified with/without default space/version inconsistently.

            try:
                container_ref = ContainerReference.model_validate(self._create_container_ref(prop.container))
            except ValidationError:
                # Error will be reported when reading the containers
                continue
            if container_ref in container_ref_by_entity:
                container_ref_entity = container_ref_by_entity[container_ref]
                read.indices[(container_ref_entity, index.suffix)].append(
                    ReadIndex(
                        prop_id=prop.container_property,
                        order=order,
                        row_no=row_no,
                        index_id=index.suffix,
                        index=created,
                    )
                )
            else:
                # Error is reported when reading the property.
                ...

    def _read_order(self, properties: dict[str, Any], loc: tuple[str | int, ...]) -> int | None:
        if "order" not in properties:
            return None
        try:
            return int(properties["order"])
        except ValueError:
            self.errors.append(
                ModelSyntaxError(
                    message=f"In {self.source.location(loc)} invalid order value '{properties['order']}'. "
                    "Must be an integer."
                )
            )
            return None

    @staticmethod
    def read_index(index: ParsedEntity, prop_id: str) -> dict[str, Any]:
        return {
            "indexType": index.prefix,
            "properties": [prop_id],
            **index.properties,
        }

    def _process_constraint(
        self,
        prop: DMSProperty,
        read: ReadProperties,
        row_no: int,
        container_ref_by_entity: dict[ContainerReference, ParsedEntity],
        container_entities: set[ParsedEntity],
    ) -> None:
        if prop.constraint is None or prop.container_property is None or prop.container is None:
            return
        loc = (self.Sheets.properties, row_no, self.PropertyColumns.constraint)
        for constraint in prop.constraint:
            data = self.read_property_constraint(constraint, prop.container_property)
            created = self._validate_adapter(ConstraintAdapter, data, loc)
            if created is None:
                continue
            order = self._read_order(constraint.properties, loc)

            if prop.container in container_entities:
                read.constraints[(prop.container, constraint.suffix)].append(
                    ReadConstraint(
                        prop_id=prop.container_property,
                        order=order,
                        constraint_id=constraint.suffix,
                        row_no=row_no,
                        constraint=created,
                    )
                )
                continue
            # The container entity was not found in the containers table. This could either be because the
            # container is missing, or because either the container entity in the Properties table and the
            # Container table are specified with/without default space/version inconsistently.
            try:
                container_ref = ContainerReference.model_validate(self._create_container_ref(prop.container))
            except ValidationError:
                # Error will be reported when reading the containers
                continue
            if container_ref in container_ref_by_entity:
                container_ref_entity = container_ref_by_entity[container_ref]
                read.constraints[(container_ref_entity, constraint.suffix)].append(
                    ReadConstraint(
                        prop_id=prop.container_property,
                        order=order,
                        constraint_id=constraint.suffix,
                        row_no=row_no,
                        constraint=created,
                    )
                )
            else:
                # Error is reported when reading the property.
                ...

    @staticmethod
    def read_property_constraint(constraint: ParsedEntity, prop_id: str) -> dict[str, Any]:
        return {"constraintType": constraint.prefix, "properties": [prop_id], **constraint.properties}

    def read_view_property(self, prop: DMSProperty, loc: tuple[str | int, ...]) -> dict[str, Any]:
        """Reads a single view property from a given row in the properties table.

        The type of property (core, edge, reverse direct relation) is determined based on the connection column
        as follows:
        1. If the connection is empty or 'direct' it is a core property.
        2. If the connection is 'edge' it is an edge property.
        3. If the connection is 'reverse' it is a reverse direct relation property
        4. Otherwise, it is an error.

        Args:
            prop (DMSProperty): The property row to read.
            loc (tuple[str | int, ...]): The location of the property in the source for error reporting.

        Returns:
            ViewRequestProperty: The parsed view property.
        """

        if prop.connection is None or prop.connection.suffix == "direct":
            return self.read_core_view_property(prop)
        elif prop.connection.suffix == "edge":
            return self.read_edge_view_property(prop, loc)
        elif prop.connection.suffix == "reverse":
            return self.read_reverse_direct_relation_view_property(prop)
        else:
            self.errors.append(
                ModelSyntaxError(
                    message=f"In {self.source.location(loc)} invalid connection type '{prop.connection.suffix}'. "
                )
            )
            return {}

    def read_core_view_property(self, prop: DMSProperty) -> dict[str, Any]:
        source: dict[str, str | None] | None = None
        if prop.connection is not None and prop.value_type.suffix != self.CELL_MISSING:
            source = self._create_view_ref(prop.value_type)

        return dict(
            connectionType="primary_property",
            name=prop.name,
            description=prop.description,
            container=self._create_container_ref(prop.container),
            containerPropertyIdentifier=prop.container_property,
            source=source,
        )

    def read_edge_view_property(self, prop: DMSProperty, loc: tuple[str | int, ...]) -> dict[str, Any]:
        if prop.connection is None:
            return {}
        edge_source: dict[str, str | None] | None = None
        if "edgeSource" in prop.connection.properties:
            edge_source = self._create_view_ref_unparsed(
                prop.connection.properties["edgeSource"], (*loc, self.PropertyColumns.connection, "edgeSource")
            )
        return dict(
            connectionType="single_edge_connection" if prop.max_count == 1 else "multi_edge_connection",
            name=prop.name,
            description=prop.description,
            source=self._create_view_ref(prop.value_type),
            type=self._create_node_ref_unparsed(
                prop.connection.properties.get("type"),
                prop.view,
                prop.view_property,
                (*loc, self.PropertyColumns.connection, "type"),
            ),
            edgeSource=edge_source,
            direction=prop.connection.properties.get("direction", "outwards"),
        )

    def read_reverse_direct_relation_view_property(
        self,
        prop: DMSProperty,
    ) -> dict[str, Any]:
        if prop.connection is None:
            return {}
        view_ref = self._create_view_ref(prop.value_type)
        return dict(
            connectionType="single_reverse_direct_relation" if prop.max_count == 1 else "multi_reverse_direct_relation",
            name=prop.name,
            description=prop.description,
            source=view_ref,
            through={
                "source": view_ref,
                "identifier": prop.connection.properties.get("property"),
            },
        )

    def read_container_property(
        self, prop: DMSProperty, enum_collections: dict[str, dict[str, Any]], loc: tuple[str | int, ...]
    ) -> dict[str, Any]:
        data_type = self._read_data_type(prop, enum_collections, loc)
        return dict(
            immutable=prop.immutable,
            nullable=prop.min_count == 0 or prop.min_count is None,
            autoIncrement=prop.auto_increment,
            defaultValue=prop.default,
            description=prop.container_property_description,
            name=prop.container_property_name,
            type=data_type,
        )

    def _read_data_type(
        self, prop: DMSProperty, enum_collections: dict[str, dict[str, Any]], loc: tuple[str | int, ...]
    ) -> dict[str, Any]:
        # Implementation to read the container property type from DMSProperty
        is_list = None if prop.max_count is None else prop.max_count > 1
        max_list_size: int | None = None
        if is_list and prop.max_count is not None:
            max_list_size = prop.max_count

        args: dict[str, Any] = {
            "maxListSize": max_list_size,
            "list": is_list,
            "type": "direct" if prop.connection is not None else prop.value_type.suffix,
        }
        args.update(prop.value_type.properties)
        if "container" in args and prop.connection is not None:
            # Direct relation constraint.
            args["container"] = self._create_container_ref_unparsed(
                prop.connection.properties["container"], (*loc, self.PropertyColumns.connection, "container")
            )
        if args["type"] == "enum" and "collection" in prop.value_type.properties:
            args["values"] = enum_collections.get(prop.value_type.properties["collection"], {})
        return args

    def read_containers(
        self, containers: list[DMSContainer], properties: ProcessedProperties
    ) -> list[ContainerRequest]:
        # Implementation to read containers from DMSContainer list
        containers_requests: list[ContainerRequest] = []
        rows_by_seen: dict[ParsedEntity, list[int]] = defaultdict(list)
        for row_no, container in enumerate(containers):
            property_constraints = properties.constraints.get(container.container, {})
            require_constraints = self.read_container_constraints(container, row_no)
            if conflict := set(property_constraints.keys()).intersection(set(require_constraints.keys())):
                conflict_str = humanize_collection(conflict)
                location_str = self.source.location((self.Sheets.containers, row_no, self.ContainerColumns.constraint))
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {location_str} the container '{container.container!s}' has constraints defined "
                            f"with the same identifier(s) as the uniqueness constraint defined in the "
                            f"{self.Sheets.properties} sheet. Ensure that the identifiers are unique. "
                            f"Conflicting identifiers: {conflict_str}. "
                        )
                    )
                )
            constraints = {**property_constraints, **require_constraints}
            container_request = self._validate_obj(
                ContainerRequest,
                dict(
                    **self._create_container_ref(container.container),
                    usedFor=container.used_for,
                    name=container.name,
                    description=container.description,
                    properties=properties.container[container.container],
                    indexes=properties.indices.get(container.container),
                    constraints=constraints or None,
                ),
                (self.Sheets.containers, row_no),
            )
            if container_request is None:
                continue
            if container.container in rows_by_seen:
                rows_by_seen[container.container].append(row_no)
            else:
                containers_requests.append(container_request)
                rows_by_seen[container.container] = [row_no]
        for entity, rows in rows_by_seen.items():
            if len(rows) > 1:
                rows_str = humanize_collection([self.source.adjust_row_number(self.Sheets.containers, r) for r in rows])
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location((self.Sheets.containers,))} the values in "
                            f"column {self.ContainerColumns.container!r} must be unique. "
                            f"Duplicated entries for container '{entity!s}' found in rows {rows_str}."
                        )
                    )
                )
        return containers_requests

    def read_container_constraints(self, container: DMSContainer, row_no: int) -> dict[str, Constraint]:
        constraints: dict[str, Constraint] = {}
        if not container.constraint:
            return constraints
        for entity in container.constraint:
            loc = self.Sheets.containers, row_no, self.ContainerColumns.constraint
            if entity.prefix != "requires":
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location(loc)} the constraint '{entity.suffix}' on container "
                            f"'{container.container!s}' has an invalid type '{entity.prefix}'. Only 'requires' "
                            f"constraints are supported at the container level."
                        )
                    )
                )
                continue

            if "require" not in entity.properties:
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location(loc)} the constraint '{entity.suffix}' on container "
                            f"'{container.container!s}' is missing the "
                            f"'require' property which is required for container level constraints."
                        )
                    )
                )
                continue
            data = {
                "constraintType": entity.prefix,
                "require": self._create_container_ref_unparsed(entity.properties["require"], loc),
            }
            created = self._validate_adapter(ConstraintAdapter, data, loc)
            if created is None:
                continue
            constraints[entity.suffix] = created
        return constraints

    def read_views(
        self,
        views: list[DMSView],
        properties: dict[ParsedEntity, dict[str, ViewRequestProperty]],
    ) -> tuple[list[ViewRequest], set[ParsedEntity]]:
        views_requests: list[ViewRequest] = []
        rows_by_seen: dict[ParsedEntity, list[int]] = defaultdict(list)
        for row_no, view in enumerate(views):
            view_request = self._validate_obj(
                ViewRequest,
                dict(
                    **self._create_view_ref(view.view),
                    name=view.name,
                    description=view.description,
                    implements=[self._create_view_ref(impl) for impl in view.implements] if view.implements else None,
                    filter=self._create_filter_dict(view.filter, row_no) if view.filter else None,
                    properties=properties.get(view.view, {}),
                ),
                (self.Sheets.views, row_no),
            )
            if view_request is None:
                continue
            if view.view in rows_by_seen:
                rows_by_seen[view.view].append(row_no)
            else:
                views_requests.append(view_request)
                rows_by_seen[view.view] = [row_no]
        for entity, rows in rows_by_seen.items():
            if len(rows) > 1:
                rows_str = humanize_collection([self.source.adjust_row_number(self.Sheets.views, r) for r in rows])
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location((self.Sheets.views,))} the values in "
                            f"column {self.ViewColumns.view!r} must be unique. "
                            f"Duplicated entries for view '{entity!s}' found in rows {rows_str}."
                        )
                    )
                )
        return views_requests, set(rows_by_seen.keys())

    def _create_filter_dict(self, filter: TableViewFilter, row_no: int) -> dict[str, Any] | None:
        if isinstance(filter, RAWFilterTableFilter):
            try:
                return json.loads(filter.filter)
            except ValueError as e:
                self.errors.append(
                    ModelSyntaxError(
                        message=(
                            f"In {self.source.location((self.Sheets.views, row_no, self.ViewColumns.filter))} "
                            f"must be valid json. Got error {e!s}"
                        )
                    )
                )
            return None
        elif isinstance(filter, EntityTableFilter):
            return self._create_entity_filter_dict(filter)
        else:
            # This is unreachable due to validation of the TableViewFilter model.
            raise RuntimeError(f"Unknown filter type {filter.__class__.__name__}")

    def _create_entity_filter_dict(self, filter: EntityTableFilter) -> dict[str, Any]:
        """Creates the filter dictionary from an EntityTableFilter."""
        if filter.type == "hasData":
            return {
                "hasData": [{**self._create_container_ref(entity), "type": "container"} for entity in filter.entities]
            }
        elif filter.type == "nodeType":
            if len(filter.entities) == 1:
                return {"equals": {"property": ["node", "type"], "value": self._create_node_ref(filter.entities[0])}}
            else:
                return {
                    "in": {
                        "property": ["node", "type"],
                        "values": [self._create_node_ref(entity) for entity in filter.entities],
                    }
                }
        else:
            # This is unreachable due to validation of the EntityTableFilter model.
            raise RuntimeError(f"Unknown filter type {filter.__class__.__name__}")

    def read_data_model(self, tables: TableDMS, valid_view_entities: set[ParsedEntity]) -> DataModelRequest:
        data: dict[str, Any] = {
            **{meta.key: meta.value for meta in tables.metadata},
            "views": [self._create_view_ref(view.view) for view in tables.views if view.view in valid_view_entities],
        }
        if description := self._create_description_field(data):
            data["description"] = description
        model = self._validate_obj(DataModelRequest, data, (self.Sheets.metadata,), field_name="value")
        if model is None:
            # This is the last step, so we can raise the error here.
            raise DataModelImportException(self.errors) from None
        return model

    def _create_description_field(self, data: dict[str, Any]) -> str | None:
        """DataModelRequest does not have a 'creator' field, this is a special addition that the Neat tables
        format supports (and recommends using). To keep it, Neat adds it to the suffix of the description field.
        """
        if CREATOR_KEY not in data and CREATOR_KEY.title() not in data:
            return None
        creator_val = data.pop(CREATOR_KEY, data.pop(CREATOR_KEY.title(), None))

        if not creator_val:
            return None

        creator = str(creator_val)
        # We do a split/join to clean up any spaces around commas. Ensuring that we have a consistent
        # canonical format.
        cleaned_creator = ", ".join(item.strip() for item in creator.split(","))
        if not cleaned_creator:
            return None
        suffix = f"{CREATOR_MARKER}{cleaned_creator}"
        description = data.get("description", "")
        if len(description) + len(suffix) > DATA_MODEL_DESCRIPTION_MAX_LENGTH:
            description = description[: DATA_MODEL_DESCRIPTION_MAX_LENGTH - len(suffix) - 4] + "..."
        if description:
            description = f"{description} {suffix}"
        else:
            description = suffix
        return description

    def _parse_entity(self, entity: str, loc: tuple[str | int, ...]) -> ParsedEntity | None:
        try:
            parsed = parse_entity(entity)
        except ValueError as e:
            self.errors.append(
                ModelSyntaxError(message=f"In {self.source.location(loc)} failed to parse entity '{entity}': {e!s}")
            )
            return None
        return parsed

    def _create_view_ref_unparsed(self, entity: str, loc: tuple[str | int, ...]) -> dict[str, str | None]:
        parsed = self._parse_entity(entity, loc)
        if parsed is None:
            return dict()
        return self._create_view_ref(parsed)

    def _create_view_ref(self, entity: ParsedEntity | None) -> dict[str, str | None]:
        if entity is None or entity.suffix == "":
            # If no suffix is given, we cannot create a valid reference.
            return dict()
        space, version = entity.prefix, entity.properties.get("version")
        if space == "":
            space = self.default_space
            # Only if default space is used, we can use default version.
            if version is None:
                version = self.default_version
        return {
            "space": space,
            "externalId": entity.suffix,
            "version": version,
        }

    def _create_container_ref_unparsed(self, entity: str, loc: tuple[str | int, ...]) -> dict[str, str]:
        parsed = self._parse_entity(entity, loc)
        if parsed is None:
            return dict()
        return self._create_container_ref(parsed)

    def _create_container_ref(self, entity: ParsedEntity | None) -> dict[str, str]:
        if entity is None or entity.suffix == "":
            # If no suffix is given, we cannot create a valid reference.
            return dict()
        return {
            "space": entity.prefix or self.default_space,
            "externalId": entity.suffix,
        }

    def _create_node_ref_unparsed(
        self, entity: str | None, view: ParsedEntity, view_prop: str, loc: tuple[str | int, ...]
    ) -> dict[str, str | None]:
        if entity is None:
            # Use default
            return self._create_node_ref(None, view, view_prop)
        parsed = self._parse_entity(entity, loc)
        if parsed is None:
            return dict()
        return self._create_node_ref(parsed, view, view_prop)

    @overload
    def _create_node_ref(
        self, entity: ParsedEntity, *, view: None = None, view_prop: None = None
    ) -> dict[str, str | None]: ...

    @overload
    def _create_node_ref(
        self, entity: ParsedEntity | None, view: ParsedEntity, view_prop: str
    ) -> dict[str, str | None]: ...

    def _create_node_ref(
        self, entity: ParsedEntity | None, view: ParsedEntity | None = None, view_prop: str | None = None
    ) -> dict[str, str | None]:
        if entity is None or entity.suffix == "":
            if view is None or view_prop is None:
                return dict()
            # If no suffix is given, we fallback to the view's property
            return {
                "space": view.prefix or self.default_space,
                "externalId": f"{view.suffix}.{view_prop}",
            }
        return {
            "space": entity.prefix or self.default_space,
            "externalId": entity.suffix,
        }

    def _validate_obj(
        self,
        obj: type[T_BaseModel],
        data: dict,
        parent_loc: tuple[str | int, ...],
        field_name: Literal["field", "column", "value"] = "column",
    ) -> T_BaseModel | None:
        try:
            return obj.model_validate(data)
        except ValidationError as e:
            self._add_error_messages(e, parent_loc, field_name=field_name)
            return None

    def _validate_adapter(
        self, adapter: TypeAdapter[T_BaseModel], data: dict[str, Any], parent_loc: tuple[str | int, ...]
    ) -> T_BaseModel | None:
        try:
            return adapter.validate_python(data, strict=True)
        except ValidationError as e:
            self._add_error_messages(e, parent_loc, field_name="column")
            return None

    def _add_error_messages(
        self,
        error: ValidationError,
        parent_loc: tuple[str | int, ...],
        field_name: Literal["field", "column", "value"] = "column",
    ) -> None:
        context = ValidationContext(
            parent_loc=parent_loc,
            humanize_location=self.source.location,
            field_name=field_name,
            missing_required_descriptor="empty" if field_name == "column" else "missing",
        )

        if field_renaming := self.source.field_mapping(parent_loc[0]):
            context.field_renaming = field_renaming

        seen: set[str] = set()
        for error_details in error.errors(include_input=True, include_url=False):
            message = humanize_validation_error(error_details, context)
            if message in seen:
                continue

            seen.add(message)
            self.errors.append(ModelSyntaxError(message=message))



from abc import ABC
from typing import TYPE_CHECKING

from cognite.neat._issues import ConsistencyError

if TYPE_CHECKING:
    from cognite.neat._issues import ModelSyntaxError
    from cognite.neat._utils.http_client import HTTPMessage


class NeatException(Exception, ABC):
    """Base class for all exceptions raised by Neat."""

    pass


class DataModelImportException(NeatException):
    """Raised when there is an error importing a model."""

    def __init__(self, errors: "list[ModelSyntaxError]") -> None:
        super().__init__(errors)
        self.errors = errors

    def __str__(self) -> str:
        return f"Model import failed with {len(self.errors)} errors: " + "; ".join(map(str, self.errors))


class DataModelCreateException(NeatException):
    """Raised when there is an error creating a data model."""

    def __init__(self, errors: "list[ModelSyntaxError | ConsistencyError]") -> None:
        super().__init__(errors)
        self.errors = errors

    def __str__(self) -> str:
        return f"Model creation failed with {len(self.errors)} errors: " + "; ".join(map(str, self.errors))


class CDFAPIException(NeatException):
    """Raised when there is an error in an API call."""

    def __init__(self, messages: "list[HTTPMessage]") -> None:
        self.messages = messages

    def __str__(self) -> str:
        return f"{type(self).__name__}: " + "; ".join(map(str, self.messages))


class FileReadException(NeatException):
    """Raised when there is an error reading a file."""

    def __init__(self, filepath: str, message: str) -> None:
        super().__init__(f"Error reading file {filepath}: {message}")
        self.filepath = filepath
        self.message = message

    def __str__(self) -> str:
        return f"Error reading file {self.filepath}: {self.message}"


class UserInputError(NeatException):
    """Raised when there is an error in user input."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return f"User input error: {self.message}"


class NeatImportError(NeatException, ImportError):
    def __init__(self, module: str, neat_extra: str, functionality: str = "functionality") -> None:
        message = (
            f"The {functionality} requires {module}. You can include it in your neat "
            f'installation with `pip install "cognite-neat[{neat_extra}]"`.'
        )
        super().__init__(message)

    def __str__(self) -> str:
        return self.args[0]



import sys
from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import MutableSequence, Sequence
from typing import Generic, Literal, TypeAlias, TypeVar

import httpx
from pydantic import BaseModel, ConfigDict, Field, JsonValue, ValidationError, model_serializer

from cognite.neat._exceptions import CDFAPIException
from cognite.neat._utils.http_client._tracker import ItemsRequestTracker
from cognite.neat._utils.useful_types import PrimaryTypes, ReferenceObject, T_Reference

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

StatusCode: TypeAlias = int


class HTTPMessage(BaseModel):
    """Base class for HTTP messages (requests and responses)"""


class FailedRequestMessage(HTTPMessage):
    message: str

    def __str__(self) -> str:
        return self.message


class ResponseMessage(HTTPMessage):
    code: StatusCode
    body: str


class SuccessResponse(ResponseMessage): ...


class ErrorDetails(BaseModel):
    """This is the structure of failure responses from CDF APIs"""

    code: StatusCode
    message: str
    missing: list[JsonValue] | None = None
    duplicated: list[JsonValue] | None = None
    is_auto_retryable: bool | None = Field(None, alias="isAutoRetryable")

    @classmethod
    def from_response(cls, response: httpx.Response) -> "ErrorDetails":
        try:
            return _ErrorResponse.model_validate_json(response.text).error
        except ValidationError:
            return cls(code=response.status_code, message=response.text)


class _ErrorResponse(BaseModel):
    error: ErrorDetails


class FailedResponse(ResponseMessage):
    error: ErrorDetails

    def __str__(self) -> str:
        return f"HTTP {self.code} | {self.error.message}"


class RequestMessage(HTTPMessage, ABC):
    """Base class for HTTP request messages"""

    endpoint_url: str
    method: Literal["GET", "POST", "PATCH", "DELETE"]
    connect_attempt: int = 0
    read_attempt: int = 0
    status_attempt: int = 0
    api_version: str | None = None

    @property
    def total_attempts(self) -> int:
        return self.connect_attempt + self.read_attempt + self.status_attempt

    @abstractmethod
    def create_success_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        raise NotImplementedError()

    @abstractmethod
    def create_failure_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        raise NotImplementedError()

    @abstractmethod
    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        raise NotImplementedError()


class SimpleRequest(RequestMessage):
    """Base class for requests with a simple success/fail response structure"""

    def create_success_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        return [SuccessResponse(code=response.status_code, body=response.text)]

    def create_failure_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        return [
            FailedResponse(code=response.status_code, body=response.text, error=ErrorDetails.from_response(response))
        ]

    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        return [FailedRequestMessage(message=error_message)]


class ParametersRequest(SimpleRequest):
    """Base class for HTTP request messages with query parameters"""

    parameters: dict[str, PrimaryTypes] | None = None


class BodyRequest(ParametersRequest, ABC):
    """Base class for HTTP request messages with a body"""

    @abstractmethod
    def data(self) -> str | bytes:
        raise NotImplementedError()


class SimpleBodyRequest(BodyRequest):
    body: str | bytes

    def data(self) -> str | bytes:
        return self.body


class ItemMessage(BaseModel, Generic[T_Reference], ABC):
    """Base class for message related to a specific item"""

    ids: Sequence[T_Reference]


class SuccessResponseItems(ItemMessage[T_Reference], SuccessResponse): ...


class FailedResponseItems(ItemMessage[T_Reference], FailedResponse): ...


class FailedRequestItems(ItemMessage[T_Reference], FailedRequestMessage): ...


T_BaseModel = TypeVar("T_BaseModel", bound=BaseModel)


class ItemBody(BaseModel, Generic[T_Reference, T_BaseModel], ABC):
    items: Sequence[T_BaseModel]
    extra_args: dict[str, JsonValue] | None = None

    @model_serializer(mode="plain", return_type=dict)
    def serialize(self) -> dict[str, JsonValue]:
        data: dict[str, JsonValue] = {
            "items": [item.model_dump(exclude_unset=False, by_alias=True, exclude_none=False) for item in self.items]
        }
        if isinstance(self.extra_args, dict):
            data.update(self.extra_args)
        return data

    @abstractmethod
    def as_ids(self) -> list[T_Reference]:
        """Returns the list of item identifiers for the items in the body."""
        raise NotImplementedError()

    def split(self, mid: int) -> tuple[Self, Self]:
        """Splits the body into two smaller bodies.

        This is useful for retrying requests that fail due to size limits or timeouts.

        Args:
            mid: The index at which to split the items.
        Returns:
            A tuple containing two new ItemBody instances, each with half of the original items.

        """

        type_ = type(self)
        return type_(items=self.items[:mid], extra_args=self.extra_args), type_(
            items=self.items[mid:], extra_args=self.extra_args
        )


class ItemIDBody(ItemBody[ReferenceObject, ReferenceObject]):
    def as_ids(self) -> list[ReferenceObject]:
        return list(self.items)


class ItemsRequest(BodyRequest, Generic[T_Reference, T_BaseModel]):
    """Requests message for endpoints that accept multiple items in a single request.

    This class provides functionality to split large requests into smaller ones, handle responses for each item,
    and manage errors effectively.

    Attributes:
        body (ItemBody): The body of the request containing the items to be processed.
        max_failures_before_abort (int): The maximum number of failed split requests before aborting further splits.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    body: ItemBody[T_Reference, T_BaseModel]
    max_failures_before_abort: int = 50
    tracker: ItemsRequestTracker | None = None

    def data(self) -> str:
        return self.body.model_dump_json(exclude_unset=True, by_alias=True)

    def split(self, status_attempts: int) -> "list[ItemsRequest]":
        """Splits the request into two smaller requests.

        This is useful for retrying requests that fail due to size limits or timeouts.

        Args:
            status_attempts: The number of status attempts to set for the new requests. This is used when the
                request failed with a 5xx status code and we want to track the number of attempts. For 4xx errors,
                there is at least one item causing the error, so we do not increment the status attempts, but
                instead essentially do a binary search to find the problematic item(s).

        Returns:
            A list containing two new ItemsRequest instances, each with half of the original items.

        """
        mid = len(self.body.items) // 2
        if mid == 0:
            return [self]
        tracker = self.tracker or ItemsRequestTracker(self.max_failures_before_abort)
        tracker.register_failure()
        messages: list[ItemsRequest] = []
        for body in self.body.split(mid):
            item_request = ItemsRequest(
                endpoint_url=self.endpoint_url,
                method=self.method,
                body=body,
                connect_attempt=self.connect_attempt,
                read_attempt=self.read_attempt,
                status_attempt=status_attempts,
            )
            item_request.tracker = tracker
            messages.append(item_request)
        return messages

    def create_success_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        return [SuccessResponseItems(code=response.status_code, body=response.text, ids=self.body.as_ids())]

    def create_failure_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        """Creates response messages based on the HTTP response and the original request.

        Args:
            response: The HTTP response received from the server.
        Returns:
            A sequence of HTTPMessage instances representing the outcome for each item in the request.
        """
        return [
            FailedResponseItems(
                code=response.status_code,
                body=response.text,
                error=ErrorDetails.from_response(response),
                ids=self.body.as_ids(),
            )
        ]

    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        """Creates failed request messages for each item in the request.

        Args:
            error_message: The error message to include in the failed request messages.

        Returns:
            A sequence of HTTPMessage instances representing the failed request for each item.
        """
        return [FailedRequestItems(message=error_message, ids=self.body.as_ids())]


class APIResponse(UserList, MutableSequence[ResponseMessage | FailedRequestMessage]):
    def __init__(self, collection: Sequence[ResponseMessage | FailedRequestMessage] | None = None):
        super().__init__(collection or [])

    def raise_for_status(self) -> None:
        error_messages = [message for message in self.data if not isinstance(message, SuccessResponse)]
        if error_messages:
            raise CDFAPIException(error_messages)

    @property
    def success_response(self) -> SuccessResponse:
        success = [msg for msg in self.data if isinstance(msg, SuccessResponse)]
        if len(success) == 1:
            return success[0]
        elif success:
            raise ValueError("Multiple successful HTTP responses found in the messages.")
        else:
            raise ValueError("No successful HTTP response found in the messages.")



from typing import Any

from pydantic import Field, ValidationInfo, field_validator, model_validator

from cognite.neat._data_model.models.entities import URI, ConceptEntity, DataType, UnknownEntity

from ._base import ResourceMetadata


class Property(ResourceMetadata):
    value_type: DataType | ConceptEntity | UnknownEntity = Field(
        union_mode="left_to_right",
        description="Value type that the property can hold. It takes either subset of XSD type or a class defined.",
    )
    min_count: int | None = Field(
        default=None,
        ge=0,
        description="Minimum number of values that the property can hold. "
        "If no value is provided, the default value is None meaning `0`, "
        "which means that the property is optional.",
    )
    max_count: int | None = Field(
        default=None,
        ge=0,
        description="Maximum number of values that the property can hold. "
        "If no value is provided, the default value is None meaning `inf`, "
        "which means that the property can hold any number of values (listable).",
    )
    default: Any | None = Field(alias="Default", default=None, description="Default value of the property.")
    instance_reference: list[URI] | None = Field(
        default=None,
        description="The URI(s) in the graph to get the value of the property.",
    )

    @model_validator(mode="after")
    def check_min_max_count(self) -> "Property":
        if self.min_count is not None and self.max_count is not None:
            if self.min_count > self.max_count:
                raise ValueError("min_count must be less than or equal to max_count")
        return self

    @field_validator("default", mode="after")
    def check_default_value_primitive_type(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value

        value_type = info.data.get("value_type")
        if not isinstance(value_type, DataType):
            raise ValueError("Setting default value is only supported for primitive value types.")
        return value

    @field_validator("default", mode="after")
    def check_default_value_python_type_exists(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value

        value_type = info.data.get("value_type")

        if isinstance(value_type, DataType) and not hasattr(value_type, "python"):
            raise ValueError(
                f"DataType {value_type} does not have a python type defined."
                " Setting default value for property is not possible."
            )
        return value

    @field_validator("default", mode="after")
    def check_default_value_not_list(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value

        if isinstance(value, list):
            raise ValueError("Setting list as default value is not supported.")
        return value

    @field_validator("default", mode="after")
    def check_default_value_single_valued(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value

        max_count = info.data.get("max_count")

        if max_count is None or max_count > 1:
            raise ValueError(
                "Setting default value is only supported for single-valued properties."
                f" Property has max_count={max_count or 'Inf'}."
            )
        return value

    @field_validator("default", mode="after")
    def check_default_value_type_match(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value

        value_type = info.data.get("value_type")

        if (
            isinstance(value_type, DataType)
            and hasattr(value_type, "python")
            and not isinstance(value, value_type.python)
        ):
            raise ValueError(
                f"Default value type is {type(value)}, which does not match expected value type {value_type.python}."
            )
        return value



from typing import Literal

from pydantic import Field

from cognite.neat._utils.useful_types import ReferenceObject

from ._constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    INSTANCE_ID_PATTERN,
    SPACE_FORMAT_PATTERN,
)


class SpaceReference(ReferenceObject):
    space: str = Field(
        description="Id of the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )


class ContainerReference(ReferenceObject):
    type: Literal["container"] = Field("container", exclude=True)
    space: str = Field(
        description="Id of the space hosting (containing) the container.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"


class ViewReference(ReferenceObject):
    type: Literal["view"] = Field("view", exclude=True)
    space: str = Field(
        description="Id of the space that the view belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the view.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the view.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(version={self.version})"


class DataModelReference(ReferenceObject):
    space: str = Field(
        description="Id of the space that the data model belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(version={self.version})"


class NodeReference(ReferenceObject):
    space: str = Field(
        description="Id of the space hosting (containing) the node.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the node.",
        min_length=1,
        max_length=255,
        pattern=INSTANCE_ID_PATTERN,
    )


class ContainerDirectReference(ReferenceObject):
    source: ContainerReference = Field(
        description="Reference to the container from where this relation is inherited.",
    )
    identifier: str = Field(
        description="Identifier of the relation in the source container.",
        min_length=1,
        max_length=255,
        pattern=CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    )

    def __str__(self) -> str:
        return f"{self.source!s}.{self.identifier}"


class ViewDirectReference(ReferenceObject):
    source: ViewReference = Field(
        description="Reference to the view from where this relation is inherited.",
    )
    identifier: str = Field(
        description="Identifier of the relation in the source view.",
        min_length=1,
        max_length=255,
        pattern=CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    )

    def __str__(self) -> str:
        return f"{self.source!s}.{self.identifier}"


class ContainerIndexReference(ReferenceObject):
    """Reference to a container index for deletion API."""

    space: str = Field(
        description="Id of the space hosting the container.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    container_external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
        alias="containerExternalId",
    )
    identifier: str = Field(
        description="Identifier of the index.",
    )


class ContainerConstraintReference(ReferenceObject):
    """Reference to a container constraint for deletion API."""

    space: str = Field(
        description="Id of the space hosting the container.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    container_external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
        alias="containerExternalId",
    )
    identifier: str = Field(
        description="Identifier of the constraint.",
    )



from ._base import DataModelTransformer, VerifiedDataModelTransformer
from ._converters import (
    AddCogniteProperties,
    AddConceptImplements,
    ChangeViewPrefix,
    ClassicPrepareCore,
    ConceptualToPhysical,
    ConversionTransformer,
    ConvertToDataModel,
    DropModelViews,
    IncludeReferenced,
    MergeConceptualDataModels,
    MergePhysicalDataModels,
    PhysicalToConceptual,
    PrefixEntities,
    SetIDDMSModel,
    StandardizeNaming,
    StandardizeSpaceAndVersion,
    SubsetConceptualDataModel,
    SubsetPhysicalDataModel,
    ToCompliantEntities,
    ToDataProductModel,
    ToDMSCompliantEntities,
    ToEnterpriseModel,
    ToExtensionModel,
    ToSolutionModel,
)
from ._mapping import AsParentPropertyId, MapOneToOne, PhysicalDataModelMapper
from ._union_conceptual import UnionConceptualDataModel
from ._verification import (
    VerifyAnyDataModel,
    VerifyConceptualDataModel,
    VerifyPhysicalDataModel,
)

__all__ = [
    "AddCogniteProperties",
    "AddConceptImplements",
    "AsParentPropertyId",
    "ChangeViewPrefix",
    "ClassicPrepareCore",
    "ConceptualToPhysical",
    "ConversionTransformer",
    "ConvertToDataModel",
    "DataModelTransformer",
    "DropModelViews",
    "IncludeReferenced",
    "MapOneToOne",
    "MergeConceptualDataModels",
    "MergePhysicalDataModels",
    "PhysicalDataModelMapper",
    "PhysicalToConceptual",
    "PrefixEntities",
    "SetIDDMSModel",
    "StandardizeNaming",
    "StandardizeSpaceAndVersion",
    "SubsetConceptualDataModel",
    "SubsetPhysicalDataModel",
    "ToCompliantEntities",
    "ToDMSCompliantEntities",
    "ToDataProductModel",
    "ToEnterpriseModel",
    "ToExtensionModel",
    "ToSolutionModel",
    "UnionConceptualDataModel",
    "VerifiedDataModelTransformer",
    "VerifyAnyDataModel",
    "VerifyConceptualDataModel",
    "VerifyPhysicalDataModel",
]



import json
import uuid
from collections import defaultdict
from typing import Any

from cognite.neat._issues import ConsistencyError, IssueList, ModelSyntaxError, Recommendation
from cognite.neat._session._html._render import render
from cognite.neat._store import NeatStore


class Issues:
    """Class to handle issues in the NeatSession."""

    def __init__(self, store: NeatStore) -> None:
        self._store = store

    @property
    def _issues(self) -> IssueList:
        """Get all issues from the last change in the store."""
        issues = IssueList()
        if change := self._store.provenance.last_change:
            issues += change.errors or IssueList()
            issues += change.issues or IssueList()
        return issues

    @property
    def _stats(self) -> dict[str, Any]:
        """Compute statistics about issues."""
        by_type: defaultdict[str, int] = defaultdict(int)
        by_code: defaultdict[str, int] = defaultdict(int)

        stats: dict[str, Any] = {
            "total": len(self._issues),
            "by_type": by_type,
            "by_code": by_code,
            "severity_order": [ModelSyntaxError.__name__, ConsistencyError.__name__, Recommendation.__name__],
        }

        for issue in self._issues:
            stats["by_type"][issue.issue_type()] += 1

            if issue.code:
                stats["by_code"][f"{issue.issue_type()}:{issue.code}"] += 1

        return stats

    @property
    def _serialized_issues(self) -> list[dict[str, Any]]:
        """Convert issues to JSON-serializable format."""
        serialized = []
        for idx, issue in enumerate(self._issues):
            serialized.append(
                {
                    "id": idx,
                    "type": issue.issue_type(),
                    "code": issue.code or "",
                    "message": issue.message,
                    "fix": issue.fix or "",
                }
            )
        return serialized

    def _repr_html_(self) -> str:
        """Generate interactive HTML representation."""
        if not self._issues:
            return "<b>No issues found.</b>"
        stats = self._stats

        # Generate unique ID for this render to avoid conflicts in Jupyter
        unique_id = uuid.uuid4().hex[:8]

        template_vars = {
            "JSON": json.dumps(self._serialized_issues),
            "total": stats["total"],
            "syntax_errors": stats["by_type"].get("ModelSyntaxError", 0),
            "consistency_errors": stats["by_type"].get("ConsistencyError", 0),
            "recommendations": stats["by_type"].get("Recommendation", 0),
            "unique_id": unique_id,
        }

        return render("issues", template_vars)



from dataclasses import dataclass
from typing import Literal

from cognite.neat._v0.core._issues import NeatError


@dataclass(unsafe_hash=True)
class NeatValueError(NeatError, ValueError):
    """{raw_message}"""

    raw_message: str


@dataclass(unsafe_hash=True)
class NeatTypeError(NeatError, TypeError):
    """{raw_message}"""

    raw_message: str


@dataclass(unsafe_hash=True)
class RegexViolationError(NeatError, ValueError):
    """The value '{value}' failed regex, {regex}, validation in {location}.
    Make sure that the name follows the regex pattern."""

    value: str
    regex: str
    location: str


@dataclass(unsafe_hash=True)
class NeatImportError(NeatError, ImportError):
    """The functionality requires {module}. You can include it
    in your neat installation with `pip install "cognite-neat[{neat_extra}]"`.
    """

    module: str
    neat_extra: str


@dataclass(unsafe_hash=True)
class WillExceedLimitError(NeatError, RuntimeError):
    """Cannot write {resource_count} {resource_type} to project {project} as the current available capacity
    is {available_capacity} {resource_type}. Neat requires a capacity of at least {margin} {resource_type} are
    left for future writes, {available_capacity}-{resource_count} < {margin}."""

    resource_type: Literal["instances"]
    resource_count: int
    project: str
    available_capacity: int
    margin: int



import difflib
from pathlib import Path
from typing import Any, Literal

import yaml
from pydantic import ValidationError

from cognite.neat._client import NeatClient
from cognite.neat._data_model._analysis import ValidationResources
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.importers._base import DMSImporter
from cognite.neat._data_model.models.dms import (
    DataModelReference,
    RequestSchema,
    SpaceReference,
)
from cognite.neat._data_model.models.dms._data_model import DataModelRequest
from cognite.neat._data_model.models.dms._references import ViewReference
from cognite.neat._data_model.models.dms._views import ViewRequest
from cognite.neat._data_model.models.entities._parser import parse_entities
from cognite.neat._exceptions import (
    CDFAPIException,
    DataModelCreateException,
    DataModelImportException,
    FileReadException,
)
from cognite.neat._issues import ConsistencyError, ModelSyntaxError
from cognite.neat._utils.http_client import FailedRequestMessage
from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.validation import ValidationContext, humanize_validation_error


class DMSAPIImporter(DMSImporter):
    """Imports DMS in the API format."""

    ENCODING = "utf-8"

    def __init__(self, schema: RequestSchema | dict[str, Any]) -> None:
        self._schema = schema

    def to_data_model(self) -> RequestSchema:
        if isinstance(self._schema, RequestSchema):
            return self._schema
        try:
            return RequestSchema.model_validate(self._schema)
        except ValidationError as e:
            context = ValidationContext()
            errors = [
                ModelSyntaxError(message=humanize_validation_error(error, context))
                for error in e.errors(include_input=True, include_url=False)
            ]
            raise DataModelImportException(errors) from None

    @classmethod
    def from_cdf(cls, data_model: DataModelReference, client: NeatClient) -> "DMSAPIImporter":
        """Create a DMSAPIImporter from a data model in CDF."""
        data_models = client.data_models.retrieve([data_model])
        if not data_models:
            available_data_models = [
                str(model.as_reference()) for model in client.data_models.list(limit=1000, include_global=True)
            ]
            close_matches = difflib.get_close_matches(str(data_model), available_data_models, n=1, cutoff=0.9)
            suggestion_msg = ""
            if close_matches:
                suggestion_msg = f" Did you mean: {close_matches[0]!r}?"
            raise CDFAPIException(
                messages=[
                    FailedRequestMessage(message=f"Data model '{data_model!s}' not found in CDF.{suggestion_msg}")
                ]
            )
        data_model = data_models[0]
        views = client.views.retrieve(data_model.views or [])
        if missing_views := set(data_model.views or []) - {view.as_reference() for view in views}:
            raise CDFAPIException(
                messages=[
                    FailedRequestMessage(
                        message=f"Views {humanize_collection(missing_views)} not found in CDF "
                        f"for data model {data_model}."
                    )
                ]
            )
        container_ids = list({container for view in views for container in view.mapped_containers})
        containers = client.containers.retrieve(container_ids)
        if missing_containers := set(container_ids) - {container.as_reference() for container in containers}:
            raise CDFAPIException(
                messages=[
                    FailedRequestMessage(
                        message=f"Containers {humanize_collection(missing_containers)} not found in CDF "
                        f"for data model {data_model}."
                    )
                ]
            )
        node_types = [nt for view in views for nt in view.node_types]
        space_ids = list(
            {data_model.space}
            | {view.space for view in views}
            | {container.space for container in containers}
            | {nt.space for nt in node_types}
        )
        spaces = client.spaces.retrieve([SpaceReference(space=space_id) for space_id in space_ids])
        if missing_spaces := set(space_ids) - {space.space for space in spaces}:
            raise CDFAPIException(
                messages=[
                    FailedRequestMessage(
                        message=f"Spaces {humanize_collection(missing_spaces)} not found in CDF "
                        f"for data model {data_model}."
                    )
                ]
            )
        return DMSAPIImporter(
            RequestSchema(
                dataModel=data_model.as_request(),
                views=[view.as_request() for view in views],
                containers=[container.as_request() for container in containers],
                nodeTypes=node_types,
                spaces=[space.as_request() for space in spaces],
            )
        )

    @classmethod
    def from_yaml(cls, yaml_file: Path) -> "DMSAPIImporter":
        """Create a DMSTableImporter from a YAML file."""
        source = cls._display_name(yaml_file)
        if yaml_file.suffix.lower() in {".yaml", ".yml", ".json"}:
            return cls(yaml.safe_load(yaml_file.read_text(encoding=cls.ENCODING)))
        elif yaml_file.is_dir():
            return cls(cls._read_yaml_files(yaml_file))
        raise FileReadException(source.as_posix(), f"Unsupported file type: {source.suffix}")

    @classmethod
    def from_json(cls, json_file: Path) -> "DMSAPIImporter":
        """Create a DMSTableImporter from a JSON file."""
        return cls.from_yaml(json_file)

    @classmethod
    def _display_name(cls, filepath: Path) -> Path:
        """Get a display-friendly version of the file path."""
        cwd = Path.cwd()
        source = filepath
        if filepath.is_relative_to(cwd):
            source = filepath.relative_to(cwd)
        return source

    @classmethod
    def _read_yaml_files(cls, directory: Path) -> dict[str, Any]:
        """Read all YAML files in a directory and combine them into a single dictionary."""
        schema_data: dict[str, Any] = {}
        data_model: dict[str, Any] | None = None
        for yaml_file in directory.rglob("**/*"):
            if yaml_file.suffix.lower() not in {".yaml", ".yml", ".json"}:
                continue
            stem = yaml_file.stem.casefold()
            if stem.endswith("datamodel") and data_model is not None:
                raise FileReadException(
                    cls._display_name(directory).as_posix(),
                    "Multiple data model files found in directory.",
                )
            data = yaml.safe_load(yaml_file.read_text(encoding=cls.ENCODING))
            list_data = data if isinstance(data, list) else [data]
            if stem.endswith("datamodel"):
                data_model = data
            elif stem.endswith("container"):
                schema_data.setdefault("containers", []).extend(list_data)
            elif stem.endswith("view"):
                schema_data.setdefault("views", []).extend(list_data)
            elif stem.endswith("space"):
                schema_data.setdefault("spaces", []).extend(list_data)
            elif stem.endswith("node"):
                schema_data.setdefault("nodeTypes", []).extend(list_data)
            # Ignore other files
        if data_model is None:
            raise FileReadException(
                cls._display_name(directory).as_posix(),
                "No data model file found in directory.",
            )
        schema_data["dataModel"] = data_model
        return schema_data


class DMSAPICreator(DMSImporter):
    """Creates data model from the existing schema components in CDF"""

    def __init__(
        self,
        cdf_snapshot: SchemaSnapshot,
        space: str,
        external_id: str,
        version: str,
        views: list[str],
        name: str | None = None,
        description: str | None = None,
        kind: Literal["solution"] = "solution",
    ) -> None:
        self._resources = ValidationResources("rebuild", local=cdf_snapshot, cdf=cdf_snapshot)
        self._space = space
        self._external_id = external_id
        self._version = version
        self._name = name
        self._description = description
        self._kind = kind
        self._views = views

    def to_data_model(self) -> RequestSchema:
        if self._kind == "solution":
            return self._to_solution_model()
        else:
            raise NotImplementedError(f"Data model kind '{self._kind}' is not supported yet.")

    def _to_solution_model(self) -> RequestSchema:
        """Creates solution data model from provided views"""

        errors, view_refs = self._parse_view_references(self._views)

        if not view_refs:
            errors.append(ModelSyntaxError(message="No valid views provided to create the data model."))

        # Check if all views exist in the cdf snapshot:
        for view in view_refs:
            if not self._resources.select_view(view):
                errors.append(
                    ConsistencyError(
                        message=f"View '{view}' not found in the provided CDF snapshot. Cannot create data model."
                    )
                )

        if errors:
            raise DataModelCreateException(errors)

        # now we iterate over the views, expand them and add them to the data model
        expanded_views: list[ViewRequest] = []

        for view in view_refs:
            expanded_view = self._resources.expand_view_properties(view)
            if not expanded_view:
                raise RuntimeError(f"Failed to expand view '{view}' despite earlier checks. This is a bug.")

            # make a deep copy to avoid modifying the cached view in resources
            expanded_view = expanded_view.model_copy(deep=True)

            # remove implements after expansion
            expanded_view.implements = None

            # set space and version to data model values
            expanded_view.space = self._space
            expanded_view.version = self._version

            # Connections should be dropped if their source are not part of the expanded view
            properties_to_remove = []
            for id, property_ in expanded_view.properties.items():
                # Currently only supporting explicit connections where source is defined
                if self._resources.is_explicit_connection(property_):
                    source = property_.source
                    if source not in view_refs:
                        properties_to_remove.append(id)

                    # Update source to point to view of solution data model
                    else:
                        property_.source = ViewReference(
                            space=self._space, external_id=source.external_id, version=self._version
                        )

            # Dropping connections which sources will not be part of the data model
            for property_id in properties_to_remove:
                expanded_view.properties.pop(property_id)

            expanded_views.append(expanded_view)

        data_model = DataModelRequest(
            space=self._space,
            externalId=self._external_id,
            version=self._version,
            name=self._name,
            description=self._description,
            views=[view.as_reference() for view in expanded_views],
        )

        return RequestSchema(dataModel=data_model, views=expanded_views)

    @classmethod
    def _parse_view_references(
        cls, views: list[str]
    ) -> tuple[list[ModelSyntaxError | ConsistencyError], list[ViewReference]]:
        """Parses list of view strings in to list of view references

        Args:
            views : list of view references in a short string format

        Returns:
            Tuple of errors and view references
        """
        view_references: list[ViewReference] = []
        errors: list[ModelSyntaxError | ConsistencyError] = []

        for view in views:
            try:
                entity = parse_entities(view)
                if not entity:
                    errors.append(ModelSyntaxError(message=f"Invalid view reference '{view}': Could not parse entity."))
                    continue

                if len(entity) != 1:
                    errors.append(
                        ModelSyntaxError(message=f"Invalid view reference '{view}': Expected a single view definition.")
                    )
                    continue

                if "version" not in entity[0].properties:
                    errors.append(
                        ModelSyntaxError(message=f"Invalid view reference '{view}': Missing 'version' property.")
                    )
                    continue

                try:
                    view_ref = ViewReference(
                        space=entity[0].prefix,
                        external_id=entity[0].suffix,
                        version=str(entity[0].properties["version"]),
                    )
                    view_references.append(view_ref)
                except ValidationError as e:
                    humanized = [humanize_validation_error(error) for error in e.errors()]
                    errors.append(
                        ModelSyntaxError(
                            message=f"Invalid view reference '{view}', cannot parse it: " + ", ".join(humanized)
                        )
                    )

            except ValueError as e:
                errors.append(ModelSyntaxError(message=f"Invalid view reference '{view}', cannot parse it: {e}"))

        return errors, view_references



from cognite.neat._data_model.models.dms._base import (
    APIResource,
    BaseModelObject,
    Resource,
    T_Resource,
    WriteableResource,
)
from cognite.neat._data_model.models.dms._constraints import (
    Constraint,
    ConstraintAdapter,
    ConstraintDefinition,
    RequiresConstraintDefinition,
    UniquenessConstraintDefinition,
)
from cognite.neat._data_model.models.dms._container import (
    Container,
    ContainerPropertyDefinition,
    ContainerRequest,
    ContainerResponse,
)
from cognite.neat._data_model.models.dms._data_types import (
    DMS_DATA_TYPES,
    BooleanProperty,
    DataType,
    DataTypeAdapter,
    DateProperty,
    DirectNodeRelation,
    EnumProperty,
    EnumValue,
    FileCDFExternalIdReference,
    Float32Property,
    Float64Property,
    FloatProperty,
    Int32Property,
    Int64Property,
    JSONProperty,
    ListablePropertyTypeDefinition,
    PropertyTypeDefinition,
    SequenceCDFExternalIdReference,
    TextProperty,
    TimeseriesCDFExternalIdReference,
    TimestampProperty,
)
from cognite.neat._data_model.models.dms._indexes import BtreeIndex, Index, IndexAdapter, IndexDefinition, InvertedIndex
from cognite.neat._data_model.models.dms._space import Space, SpaceRequest, SpaceResponse

from ._data_model import DataModelRequest, DataModelResponse
from ._http import DataModelBody, DataModelResource, ResourceId, T_DataModelResource, T_ResourceId
from ._references import (
    ContainerConstraintReference,
    ContainerDirectReference,
    ContainerIndexReference,
    ContainerReference,
    DataModelReference,
    NodeReference,
    SpaceReference,
    ViewDirectReference,
    ViewReference,
)
from ._schema import RequestSchema
from ._view_filter import (
    AVAILABLE_FILTERS,
    AndFilter,
    ContainsAllFilterData,
    ContainsAnyFilterData,
    EqualsFilterData,
    ExistsFilterData,
    Filter,
    FilterAdapter,
    FilterDataDefinition,
    HasDataFilter,
    InFilterData,
    InstanceReferencesFilterData,
    MatchAllFilterData,
    NestedFilterData,
    NotFilter,
    OrFilter,
    OverlapsFilterData,
    PrefixFilterData,
    RangeFilterData,
)
from ._view_property import (
    ConnectionPropertyDefinition,
    ConstraintOrIndexState,
    MultiEdgeProperty,
    MultiReverseDirectRelationPropertyRequest,
    MultiReverseDirectRelationPropertyResponse,
    SingleEdgeProperty,
    SingleReverseDirectRelationPropertyRequest,
    SingleReverseDirectRelationPropertyResponse,
    ViewCorePropertyRequest,
    ViewCorePropertyResponse,
    ViewPropertyDefinition,
    ViewRequestProperty,
    ViewRequestPropertyAdapter,
    ViewResponseProperty,
)
from ._views import (
    View,
    ViewRequest,
    ViewResponse,
)

__all__ = [
    "AVAILABLE_FILTERS",
    "DMS_DATA_TYPES",
    "APIResource",
    "AndFilter",
    "BaseModelObject",
    "BooleanProperty",
    "BtreeIndex",
    "ConnectionPropertyDefinition",
    "Constraint",
    "ConstraintAdapter",
    "ConstraintDefinition",
    "ConstraintOrIndexState",
    "Container",
    "ContainerConstraintReference",
    "ContainerDirectReference",
    "ContainerIndexReference",
    "ContainerPropertyDefinition",
    "ContainerReference",
    "ContainerRequest",
    "ContainerResponse",
    "ContainsAllFilterData",
    "ContainsAnyFilterData",
    "DataModelBody",
    "DataModelReference",
    "DataModelRequest",
    "DataModelResource",
    "DataModelResponse",
    "DataType",
    "DataTypeAdapter",
    "DateProperty",
    "DirectNodeRelation",
    "EnumProperty",
    "EnumValue",
    "EqualsFilterData",
    "ExistsFilterData",
    "FileCDFExternalIdReference",
    "Filter",
    "FilterAdapter",
    "FilterDataDefinition",
    "Float32Property",
    "Float64Property",
    "FloatProperty",
    "HasDataFilter",
    "InFilterData",
    "Index",
    "IndexAdapter",
    "IndexDefinition",
    "InstanceReferencesFilterData",
    "Int32Property",
    "Int64Property",
    "InvertedIndex",
    "JSONProperty",
    "ListablePropertyTypeDefinition",
    "MatchAllFilterData",
    "MultiEdgeProperty",
    "MultiReverseDirectRelationPropertyRequest",
    "MultiReverseDirectRelationPropertyResponse",
    "NestedFilterData",
    "NodeReference",
    "NotFilter",
    "OrFilter",
    "OverlapsFilterData",
    "PrefixFilterData",
    "PropertyTypeDefinition",
    "RangeFilterData",
    "RequestSchema",
    "RequiresConstraintDefinition",
    "Resource",
    "ResourceId",
    "SequenceCDFExternalIdReference",
    "SequenceCDFExternalIdReference",
    "SingleEdgeProperty",
    "SingleReverseDirectRelationPropertyRequest",
    "SingleReverseDirectRelationPropertyResponse",
    "Space",
    "SpaceReference",
    "SpaceRequest",
    "SpaceResponse",
    "T_DataModelResource",
    "T_Resource",
    "T_ResourceId",
    "TextProperty",
    "TextProperty",
    "TimeseriesCDFExternalIdReference",
    "TimeseriesCDFExternalIdReference",
    "TimestampProperty",
    "TimestampProperty",
    "UniquenessConstraintDefinition",
    "UniquenessConstraintDefinition",
    "View",
    "ViewCorePropertyRequest",
    "ViewCorePropertyResponse",
    "ViewDirectReference",
    "ViewPropertyDefinition",
    "ViewReference",
    "ViewRequest",
    "ViewRequestProperty",
    "ViewRequestPropertyAdapter",
    "ViewResponse",
    "ViewResponseProperty",
    "WriteableResource",
]



from cognite.neat._data_model.models.dms import SpaceRequest

from ._differ import ItemDiffer
from .data_classes import FieldChange


class SpaceDiffer(ItemDiffer[SpaceRequest]):
    def diff(self, cdf_space: SpaceRequest, desired_space: SpaceRequest) -> list[FieldChange]:
        return self._diff_name_description(cdf_space, desired_space)



from typing import Annotated, Any

from pydantic import BeforeValidator


def str_as_bool(value: Any) -> Any:
    if isinstance(value, str):
        val = value.lower()
        if val in {"true", "1", "yes"}:
            return True
        if val in {"false", "0", "no"}:
            return False
    # All other cases are handled by Pydantic's built-in bool validator
    return value


Bool = Annotated[bool, BeforeValidator(str_as_bool, str)]



from pathlib import Path
from typing import Any, cast

import yaml

from cognite.neat._v0.core._data_model._shared import (
    ImportedDataModel,
    T_UnverifiedDataModel,
)
from cognite.neat._v0.core._data_model.models import (
    UNVERIFIED_DATA_MODEL_BY_ROLE,
    RoleTypes,
)
from cognite.neat._v0.core._issues import IssueList, MultiValueError, NeatIssue
from cognite.neat._v0.core._issues.errors import (
    FileMissingRequiredFieldError,
    FileReadError,
)
from cognite.neat._v0.core._issues.warnings import NeatValueWarning

from ._base import BaseImporter
from ._base_file_reader import BaseFileReader


class YAMLReader:
    """Handles reading and parsing YAML files with error handling."""

    @staticmethod
    def read_file(
        filepath: Path, allowed_extensions: frozenset[str] = frozenset([".yaml", ".yml"])
    ) -> tuple[dict[str, Any], list[NeatIssue]]:
        """Read a YAML file and return the data with any issues encountered."""
        issues = BaseFileReader.validate_file(filepath, allowed_extensions)
        if issues:
            return {}, issues
        # Try to load the YAML
        try:
            data = yaml.safe_load(filepath.read_text())
            if not isinstance(data, dict):
                issues.append(FileReadError(filepath, "YAML content is not a dictionary"))
                return {}, issues
            return data, issues
        except yaml.YAMLError as exc:
            return {}, [FileReadError(filepath, f"Invalid YAML: {exc!s}")]
        except Exception as exc:
            return {}, [FileReadError(filepath, f"Error reading file: {exc!s}")]


class DictImporter(BaseImporter[T_UnverifiedDataModel]):
    """Imports the data model from a YAML file.

    Args:
        raw_data: The raw data to be imported.

    .. note::

        YAML files are typically used for storing data model when checked into version
        control systems, e.g., git-history.
        The advantage of using YAML files over Excel is that tools like git can
        show the differences between different
        versions of the data model.

    """

    def __init__(
        self,
        raw_data: dict[str, Any],
        read_issues: list[NeatIssue] | None = None,
        filepaths: list[Path] | None = None,
        source_name: str = "Unknown",
    ) -> None:
        self.raw_data = raw_data
        self._read_issues = IssueList(read_issues)
        self._filepaths = filepaths
        self._source_name = source_name

    @property
    def description(self) -> str:
        return f"YAML file {self._source_name} read as unverified data model"

    @classmethod
    def from_yaml_file(cls, filepath: Path, source_name: str = "Unknown") -> "DictImporter":
        """Create a DictImporter from a YAML file."""
        data, issues = YAMLReader.read_file(filepath)

        if issues:
            return cls({}, issues)

        return cls(data, filepaths=[filepath], source_name=source_name)

    def to_data_model(self) -> ImportedDataModel[T_UnverifiedDataModel]:
        if self._read_issues.has_errors or not self.raw_data:
            self._read_issues.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        issue_list = IssueList(title="YAML Importer", issues=self._read_issues)

        if not self._filepaths:
            issue_list.append(
                NeatValueWarning(
                    f"{type(self).__name__} was called without filepaths when there is content",
                )
            )
            metadata_file = Path()
        else:
            metadata_file_nullable = next((file for file in self._filepaths if file.stem == "metadata"), None)
            metadata_file = metadata_file_nullable or self._filepaths[0]

        if "metadata" not in self.raw_data:
            self._read_issues.append(FileMissingRequiredFieldError(metadata_file, "section", "metadata"))
            issue_list.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        metadata = self.raw_data["metadata"]

        if "role" not in metadata:
            self._read_issues.append(FileMissingRequiredFieldError(metadata, "metadata", "role"))
            issue_list.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        role_input = RoleTypes(metadata["role"])
        role_enum = RoleTypes(role_input)
        data_model_cls = UNVERIFIED_DATA_MODEL_BY_ROLE[role_enum]

        data_model = cast(T_UnverifiedDataModel, data_model_cls.load(self.raw_data))

        issue_list.trigger_warnings()
        if self._read_issues.has_errors:
            raise MultiValueError(self._read_issues.errors)

        return ImportedDataModel[T_UnverifiedDataModel](data_model)



import re
from abc import ABC
from typing import Any, Literal, TypeVar

from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic_core.core_schema import FieldSerializationInfo

from cognite.neat._utils.text import humanize_collection

from . import DirectNodeRelation
from ._base import APIResource, Resource, WriteableResource
from ._constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS,
    FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER,
    SPACE_FORMAT_PATTERN,
)
from ._references import ContainerReference, NodeReference, ViewReference
from ._view_filter import Filter
from ._view_property import (
    EdgeProperty,
    ViewCorePropertyRequest,
    ViewCorePropertyResponse,
    ViewRequestProperty,
    ViewResponseProperty,
)

KEY_PATTERN = re.compile(CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN)


class View(Resource, APIResource[ViewReference], ABC):
    space: str = Field(
        description="Id of the space that the view belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the view.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the view.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="name for the view.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the view.",
        max_length=1024,
    )
    filter: Filter | None = Field(
        default=None,
        description="A filter Domain Specific Language (DSL) used to select which instances the view should include.",
    )
    implements: list[ViewReference] | None = Field(
        default=None,
        description="References to the views from where this view will inherit properties.",
    )

    def as_reference(self) -> ViewReference:
        return ViewReference(space=self.space, external_id=self.external_id, version=self.version)

    @model_validator(mode="before")
    def set_connection_type_on_primary_properties(cls, data: dict) -> dict:
        if "properties" not in data:
            return data
        properties = data["properties"]
        if not isinstance(properties, dict):
            return data
        # We assume all properties without connectionType are core properties.
        # The reason we set connectionType it easy for pydantic to discriminate the union.
        # This also leads to better error messages, as if there is a union and pydantic do not know which
        # type to pick it will give errors from all type in the union.
        new_properties: dict[str, Any] = {}
        for prop_id, prop in properties.items():
            if isinstance(prop, dict) and "connectionType" not in prop:
                prop_copy = prop.copy()
                prop_copy["connectionType"] = "primary_property"
                new_properties[prop_id] = prop_copy
            else:
                new_properties[prop_id] = prop
        if new_properties:
            new_data = data.copy()
            new_data["properties"] = new_properties
            return new_data

        return data

    @field_validator("external_id", mode="after")
    def check_forbidden_external_id_value(cls, val: str) -> str:
        """Check the external_id not present in forbidden set"""
        if val in FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS:
            raise ValueError(
                f"'{val}' is a reserved view External ID. Reserved External IDs are: "
                f"{humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS)}"
            )
        return val

    @field_serializer("implements", mode="plain")
    @classmethod
    def serialize_implements(
        cls, implements: list[ViewReference] | None, info: FieldSerializationInfo
    ) -> list[dict[str, Any]] | None:
        if implements is None:
            return None
        output: list[dict[str, Any]] = []
        for view in implements:
            dumped = view.model_dump(**vars(info))
            dumped["type"] = "view"
            output.append(dumped)
        return output


class ViewRequest(View):
    properties: dict[str, ViewRequestProperty] = Field(
        description="View with included properties and expected edges, indexed by a unique space-local identifier."
    )

    @field_validator("properties", mode="after")
    def validate_properties_identifier(cls, val: dict[str, ViewRequestProperty]) -> dict[str, ViewRequestProperty]:
        """Validate properties Identifier"""
        return _validate_properties_keys(val)

    @property
    def used_containers(self) -> set[ContainerReference]:
        """Get all containers referenced by this view."""
        return {prop.container for prop in self.properties.values() if isinstance(prop, ViewCorePropertyRequest)}


class ViewResponse(View, WriteableResource[ViewRequest]):
    properties: dict[str, ViewResponseProperty] = Field(
        description="List of properties and connections included in this view."
    )

    created_time: int = Field(
        description="When the view was created. The number of milliseconds since 00:00:00 Thursday, 1 January 1970, "
        "Coordinated Universal Time (UTC), minus leap seconds."
    )
    last_updated_time: int = Field(
        description="When the view was last updated. The number of milliseconds since 00:00:00 Thursday, "
        "1 January 1970, Coordinated Universal Time (UTC), minus leap seconds."
    )
    writable: bool = Field(
        description="oes the view support write operations, i.e. is it writable? "
        "You can write to a view if the view maps all non-nullable properties."
    )
    queryable: bool = Field(
        description="Does the view support queries, i.e. is it queryable? You can query a view if "
        "it either has a filter or at least one property mapped to a container."
    )
    used_for: Literal["node", "edge", "all"] = Field(description="Should this operation apply to nodes, edges or both.")
    is_global: bool = Field(description="Is this a global view.")
    mapped_containers: list[ContainerReference] = Field(
        description="List of containers with properties mapped by this view."
    )

    @field_validator("properties", mode="after")
    def validate_properties_identifier(cls, val: dict[str, ViewResponseProperty]) -> dict[str, ViewResponseProperty]:
        """Validate properties Identifier"""
        return _validate_properties_keys(val)

    @property
    def node_types(self) -> list[NodeReference]:
        """Get all node types referenced by this view."""
        nodes_refs: set[NodeReference] = set()
        for prop in self.properties.values():
            if isinstance(prop, EdgeProperty):
                nodes_refs.add(prop.type)
        return list(nodes_refs)

    def as_request(self) -> ViewRequest:
        dumped = self.model_dump(by_alias=True, exclude={"properties"})
        properties: dict[str, Any] = {}
        for key, value in self.properties.items():
            if isinstance(value, ViewCorePropertyResponse) and isinstance(value.type, DirectNodeRelation):
                # Special case. In the request the source of DirectNodeRelation is set on the Property object,
                # while in the response it is set on the DirectNodeRelation object.
                request_object = value.as_request().model_dump(by_alias=True)
                request_object["source"] = value.type.source.model_dump(by_alias=True) if value.type.source else None
                properties[key] = request_object
            elif isinstance(value, WriteableResource):
                properties[key] = value.as_request().model_dump(by_alias=True)
            else:
                properties[key] = value.model_dump(by_alias=True)

        dumped["properties"] = properties
        return ViewRequest.model_validate(dumped)


T_Property = TypeVar("T_Property")


def _validate_properties_keys(properties: dict[str, T_Property]) -> dict[str, T_Property]:
    """Validate keys of a properties dictionary."""
    errors: list[str] = []
    for key in properties:
        if not KEY_PATTERN.match(key):
            errors.append(f"Property '{key}' does not match the required pattern: {KEY_PATTERN.pattern}")
        if key in FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER:
            errors.append(
                f"'{key}' is a reserved property identifier. Reserved identifiers are: "
                f"{humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER)}"
            )
    if errors:
        raise ValueError("; ".join(errors))
    return properties



import importlib.util
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from .env_vars import AVAILABLE_LOGIN_FLOWS, AVAILABLE_PROVIDERS, LoginFlow, Provider, create_env_file_content


class InteractiveFlow(ABC):
    def __init__(self, env_path: Path):
        self.env_path = env_path

    @abstractmethod
    def run(self) -> None: ...

    def create_env_file(self, provider: Provider, login_flow: LoginFlow) -> None:
        env_content = create_env_file_content(provider, login_flow)
        self.env_path.write_text(env_content, encoding="utf-8", newline="\n")


def get_interactive_flow(env_file_path: Path) -> InteractiveFlow:
    try:
        importlib.util.find_spec("IPython")
        importlib.util.find_spec("ipywidgets")
        if not _is_in_notebook():
            return NoDependencyFlow(env_file_path)
        return NotebookFlow(env_file_path)
    except ImportError:
        return NoDependencyFlow(env_file_path)


def _is_in_notebook() -> bool:
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


class NoDependencyFlow(InteractiveFlow):
    def run(self) -> None:
        if not self.should_create_env_file():
            return None
        provider = self.provider()
        login_flow: LoginFlow
        if provider != "cdf":
            login_flow = self.login_flow()
        else:
            login_flow = "client_credentials"
        self.create_env_file(provider, login_flow)
        print(f"Created environment file at {self.env_path!r}.")
        return None

    def should_create_env_file(self) -> bool:
        env_file_name = self.env_path.name
        answer = input(
            f"Would you like to create a new {env_file_name!r} file with the required environment variables? [y/N]: "
        )
        return answer.strip().lower() == "y"

    @classmethod
    def provider(cls) -> Provider:
        index = cls._prompt_choice(AVAILABLE_PROVIDERS, "Select provider:")
        return AVAILABLE_PROVIDERS[index]

    @classmethod
    def login_flow(cls) -> LoginFlow:
        index = cls._prompt_choice(AVAILABLE_LOGIN_FLOWS, "Select login flow:")
        return AVAILABLE_LOGIN_FLOWS[index]

    @classmethod
    def _prompt_choice(cls, options: tuple[str, ...], prompt: str) -> int:
        for i, option in enumerate(options, start=1):
            print(f"{i}. {option}")
        question = f"{prompt} [1-{len(options)}]: "
        while True:
            answer = input(question)
            if answer.isdigit():
                index = int(answer)
                if 1 <= index <= len(options):
                    return index - 1
            print(f"Invalid input. Please enter a number between 1 and {len(options)}.")


class NotebookFlow(InteractiveFlow):
    def __init__(self, env_path: Path):
        super().__init__(env_path)
        import ipywidgets as widgets  # type: ignore[import-untyped]
        from IPython.display import display

        self._widgets = widgets
        self._display = display

    def run(self) -> None:
        dropdown_providers = self._widgets.Dropdown(
            options=list(AVAILABLE_PROVIDERS),
            value=AVAILABLE_PROVIDERS[0],
            description="Provider:",
        )
        dropdown_login_flows = self._widgets.Dropdown(
            options=list(AVAILABLE_LOGIN_FLOWS),
            value=AVAILABLE_LOGIN_FLOWS[0],
            description="Login Flow:",
        )
        confirm_button = self._widgets.Button(description="Create template .env file", button_style="primary")
        dropdowns = self._widgets.HBox([dropdown_providers, dropdown_login_flows])
        output = self._widgets.Output()
        container = self._widgets.VBox([dropdowns, confirm_button, output])

        self._display(container)

        def on_confirm_clicked(b: Any) -> None:
            with output:
                provider = dropdown_providers.value
                login_flow = dropdown_login_flows.value
                if provider == "cdf" and login_flow != "client_credentials":
                    print(
                        "Warning: 'cdf' provider only supports 'client_credentials' login flow. Overriding selection."
                    )
                    login_flow = "client_credentials"
                is_existing = self.env_path.exists()
                self.create_env_file(provider, login_flow)
                if is_existing:
                    print(f"Overwrote existing environment file at {self.env_path!r}.")
                else:
                    print(f"Created environment file at {self.env_path!r}.")

        confirm_button.on_click(on_confirm_clicked)



from abc import ABC
from typing import Any

from pydantic import Field, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from ._base import APIResource, Resource, WriteableResource
from ._constants import (
    DATA_MODEL_DESCRIPTION_MAX_LENGTH,
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    SPACE_FORMAT_PATTERN,
)
from ._references import DataModelReference, ViewReference


class DataModel(Resource, APIResource[DataModelReference], ABC):
    """Cognite Data Model resource.

    Data models group and structure views into reusable collections.
    A data model contains a set of views where the node types can
    refer to each other with direct relations and edges.
    """

    space: str = Field(
        description="The workspace for the data model, a unique identifier for the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="Human readable name for the data model.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the data model.",
        max_length=DATA_MODEL_DESCRIPTION_MAX_LENGTH,
    )
    # The API supports View here, but in Neat we will only use ViewReference
    views: list[ViewReference] | None = Field(
        description="List of views included in this data model.",
        default=None,
    )

    def as_reference(self) -> DataModelReference:
        return DataModelReference(
            space=self.space,
            external_id=self.external_id,
            version=self.version,
        )

    @field_serializer("views", mode="plain")
    @classmethod
    def serialize_views(
        cls, views: list[ViewReference] | None, info: FieldSerializationInfo
    ) -> list[dict[str, Any]] | None:
        if views is None:
            return None
        output: list[dict[str, Any]] = []
        for view in views:
            dumped = view.model_dump(**vars(info))
            dumped["type"] = "view"
            output.append(dumped)
        return output


class DataModelRequest(DataModel): ...


class DataModelResponse(DataModel, WriteableResource[DataModelRequest]):
    created_time: int = Field(
        description="When the data model was created. The number of milliseconds since 00:00:00 Thursday, "
        "1 January 1970, Coordinated Universal Time (UTC), minus leap seconds."
    )
    last_updated_time: int = Field(
        description="When the data model was last updated. The number of milliseconds since 00:00:00 Thursday, "
        "1 January 1970, Coordinated Universal Time (UTC), minus leap seconds."
    )
    is_global: bool = Field(description="Is this a global data model.")

    def as_request(self) -> DataModelRequest:
        return DataModelRequest.model_validate(self.model_dump(by_alias=True))



from pydantic import Field

from ._base import Resource
from ._container import ContainerRequest
from ._data_model import DataModelRequest
from ._references import NodeReference
from ._space import SpaceRequest
from ._views import ViewRequest


class RequestSchema(Resource):
    """Represents a schema for creating or updating a data model in DMS."""

    data_model: DataModelRequest
    views: list[ViewRequest] = Field(default_factory=list)
    containers: list[ContainerRequest] = Field(default_factory=list)
    spaces: list[SpaceRequest] = Field(default_factory=list)
    node_types: list[NodeReference] = Field(default_factory=list)



from abc import ABC
from typing import cast

from cognite.neat._v0.core._client import NeatClient
from cognite.neat._v0.core._data_model._shared import (
    ImportedDataModel,
    T_ImportedUnverifiedDataModel,
    T_VerifiedDataModel,
    VerifiedDataModel,
)
from cognite.neat._v0.core._data_model.models import (
    ConceptualDataModel,
    PhysicalDataModel,
    UnverifiedConceptualDataModel,
    UnverifiedPhysicalDataModel,
)
from cognite.neat._v0.core._data_model.models.conceptual import ConceptualValidation
from cognite.neat._v0.core._data_model.models.physical import PhysicalValidation
from cognite.neat._v0.core._issues import MultiValueError, catch_issues
from cognite.neat._v0.core._issues.errors import NeatTypeError, NeatValueError

from ._base import DataModelTransformer


class VerificationTransformer(DataModelTransformer[T_ImportedUnverifiedDataModel, T_VerifiedDataModel], ABC):
    """Base class for all verification transformers."""

    _data_model_cls: type[T_VerifiedDataModel]
    _validation_cls: type

    def __init__(self, validate: bool = True, client: NeatClient | None = None) -> None:
        self.validate = validate
        self._client = client

    def transform(self, data_model: T_ImportedUnverifiedDataModel) -> T_VerifiedDataModel:
        in_ = data_model.unverified_data_model
        if in_ is None:
            raise NeatValueError("Cannot verify data model. The reading of the data model failed.")
        verified_data_model: T_VerifiedDataModel | None = None
        # We need to catch issues as we use the error args to provide extra context for the errors/warnings
        # For example, which row in the spreadsheet the error occurred.
        with catch_issues(data_model.context) as issues:
            data_model_cls = self._get_data_model_cls(data_model)
            dumped = in_.dump()
            verified_data_model = data_model_cls.model_validate(dumped)  # type: ignore[assignment]
            if self.validate:
                validation_cls = self._get_validation_cls(verified_data_model)  # type: ignore[arg-type]
                if issubclass(validation_cls, PhysicalValidation):
                    validation_issues = PhysicalValidation(
                        cast(PhysicalDataModel, verified_data_model),
                        self._client,
                        data_model.context,
                    ).validate()  # type: ignore[arg-type]
                elif issubclass(validation_cls, ConceptualValidation):
                    validation_issues = ConceptualValidation(verified_data_model, data_model.context).validate()  # type: ignore[arg-type]
                else:
                    raise NeatValueError("Unsupported data model type")
                issues.extend(validation_issues)

        # Raise issues which is expected to be handled outside of this method
        issues.trigger_warnings()
        if issues.has_errors:
            raise MultiValueError(issues.errors)
        if verified_data_model is None:
            raise NeatValueError("Data model was not verified")
        return verified_data_model

    def _get_data_model_cls(self, in_: T_ImportedUnverifiedDataModel) -> type[T_VerifiedDataModel]:
        return self._data_model_cls

    def _get_validation_cls(self, data_model: T_VerifiedDataModel) -> type:
        return self._validation_cls

    @property
    def description(self) -> str:
        return "Verify data model"


class VerifyPhysicalDataModel(
    VerificationTransformer[ImportedDataModel[UnverifiedPhysicalDataModel], PhysicalDataModel]
):
    """Class to verify physical data model."""

    _data_model_cls = PhysicalDataModel
    _validation_cls = PhysicalValidation

    def transform(self, data_model: ImportedDataModel[UnverifiedPhysicalDataModel]) -> PhysicalDataModel:
        return super().transform(data_model)


class VerifyConceptualDataModel(
    VerificationTransformer[ImportedDataModel[UnverifiedConceptualDataModel], ConceptualDataModel]
):
    """Class to verify conceptual data model."""

    _data_model_cls = ConceptualDataModel
    _validation_cls = ConceptualValidation

    def transform(self, data_model: ImportedDataModel[UnverifiedConceptualDataModel]) -> ConceptualDataModel:
        return super().transform(data_model)


class VerifyAnyDataModel(VerificationTransformer[T_ImportedUnverifiedDataModel, VerifiedDataModel]):
    """Class to verify arbitrary data model"""

    def _get_data_model_cls(self, in_: T_ImportedUnverifiedDataModel) -> type[VerifiedDataModel]:
        if isinstance(in_.unverified_data_model, UnverifiedConceptualDataModel):
            return ConceptualDataModel
        elif isinstance(in_.unverified_data_model, UnverifiedPhysicalDataModel):
            return PhysicalDataModel
        else:
            raise NeatTypeError(f"Unsupported data model type: {type(in_)}")

    def _get_validation_cls(self, data_model: VerifiedDataModel) -> type:
        if isinstance(data_model, ConceptualDataModel):
            return ConceptualValidation
        elif isinstance(data_model, PhysicalDataModel):
            return PhysicalValidation
        else:
            raise NeatTypeError(f"Unsupported data model type: {type(data_model)}")



import json
from pathlib import Path
from typing import cast

import yaml

from cognite.neat._data_model.exporters._base import DMSExporter, DMSFileExporter
from cognite.neat._data_model.importers._table_importer.data_classes import DMSProperty, TableDMS
from cognite.neat._data_model.models.dms import RequestSchema
from cognite.neat._utils.useful_types import DataModelTableType

from .workbook import WorkbookCreator, WorkbookOptions
from .writer import DMSTableWriter


class DMSTableExporter(DMSExporter[DataModelTableType]):
    """Exports DMS to a table structure.

    The tables can are expected to be a dictionary where the keys are the table names and the values
    are lists of dictionaries representing the rows in the table.
    """

    class Sheets:
        properties = cast(str, TableDMS.model_fields["properties"].validation_alias)

    def __init__(self, exclude_none: bool = False, skip_properties_in_other_spaces: bool = True) -> None:
        self._exclude_none = exclude_none
        self._skip_properties_in_other_spaces = skip_properties_in_other_spaces

    def export(self, data_model: RequestSchema) -> DataModelTableType:
        model = data_model.data_model
        tables = DMSTableWriter(model.space, model.version, self._skip_properties_in_other_spaces).write_tables(
            data_model
        )
        exclude: set[str] = set()
        if self._exclude_none:
            if not tables.enum:
                exclude.add("enum")
            if not tables.nodes:
                exclude.add("nodes")
            if not tables.containers:
                exclude.add("containers")

        output = tables.model_dump(mode="json", by_alias=True, exclude_none=self._exclude_none, exclude=exclude)
        # When we have exclude_none we only want to exclude none of optional properties, not required.
        # Thus, we do the implementation below
        required_properties = [
            field_.serialization_alias for field_ in DMSProperty.model_fields.values() if field_.is_required()
        ]
        for row in output[self.Sheets.properties]:
            for prop in required_properties:
                if prop not in row:
                    row[prop] = None
        return output


class DMSTableYamlExporter(DMSTableExporter, DMSFileExporter[DataModelTableType]):
    """Exports DMS to YAML."""

    def __init__(self) -> None:
        super().__init__(exclude_none=True)

    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        """Exports the data model as a flat YAML file, which is identical to the spreadsheet representation

        Args:
            data_model (RequestSchema): The data model to export.
            file_path (Path): The path to the YAML file to create.
        """
        table_format = self.export(data_model)
        file_path.write_text(
            yaml.safe_dump(table_format, sort_keys=False), encoding=self.ENCODING, newline=self.NEW_LINE
        )


class DMSTableJSONExporter(DMSTableExporter, DMSFileExporter[DataModelTableType]):
    """Exports DMS to JSON."""

    def __init__(self) -> None:
        super().__init__(exclude_none=True)

    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        """Exports the data model as a flat JSON file, which is identical to the spreadsheet representation

        Args:
            data_model (RequestSchema): The data model to export.
            file_path (Path): The path to the JSON file to create.
        """
        table_format = self.export(data_model)
        file_path.write_text(json.dumps(table_format), encoding=self.ENCODING, newline=self.NEW_LINE)


class DMSExcelExporter(DMSTableExporter, DMSFileExporter[DataModelTableType]):
    """Exports DMS to Excel file."""

    def __init__(self, options: WorkbookOptions | None = None) -> None:
        self._options = options or WorkbookOptions()
        super().__init__(
            exclude_none=False, skip_properties_in_other_spaces=self._options.skip_properties_in_other_spaces
        )

    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        """Exports the data model as a Excel file.

        Args:
            data_model (RequestSchema): The data model to export.
            file_path (Path): The path to the Excel file to create.
            options (WorkbookOptions | None): Options for creating the workbook.
        """
        table_format = self.export(data_model)
        workbook = WorkbookCreator(self._options).create_workbook(table_format)
        try:
            workbook.save(file_path)
        finally:
            workbook.close()


class DMSCsvExporter(DMSTableExporter, DMSFileExporter[DataModelTableType]):
    """Exports DMS to CSV files in a directory."""

    def export_to_file(self, data_model: RequestSchema, directory_path: Path) -> None:
        """Exports the data model as a set of CSV files, one for each table.

        Args:
            data_model (RequestSchema): The data model to export.
            directory_path (Path): The path to the directory to create the CSV files in.
        """
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import Any

from cognite.neat._data_model.deployer.data_classes import DeploymentResult
from cognite.neat._issues import IssueList


class OnSuccess(ABC):
    """Abstract base class for post-activity success handlers."""

    @abstractmethod
    def run(self, data_model: Any) -> None:
        """Execute the success handler on the data model."""
        pass


class OnSuccessIssuesChecker(OnSuccess, ABC):
    """Abstract base class for post-activity success handlers that check for issues of the data model."""

    def __init__(self) -> None:
        self._issues = IssueList()
        self._has_run = False

    @property
    def issues(self) -> IssueList:
        if not self._has_run:
            raise RuntimeError(f"{type(self).__name__} has not been run yet.")
        return IssueList(self._issues)


class OnSuccessResultProducer(OnSuccess, ABC):
    """Abstract base class for post-activity success handlers that produce desired outcomes using the data model."""

    def __init__(self) -> None:
        self._results: DeploymentResult | None = None

    @property
    def result(self) -> DeploymentResult:
        if self._results is None:
            raise RuntimeError("SchemaDeployer has not been run yet.")
        return self._results



import sys
from pathlib import Path
from typing import Any, Literal, TypeAlias

from pydantic import BaseModel, ConfigDict, Field

from cognite.neat._exceptions import UserInputError
from cognite.neat._issues import ConsistencyError, ModelSyntaxError
from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.useful_types import ModusOperandi

if sys.version_info >= (3, 11):
    import tomllib as tomli  # Python 3.11+
else:
    import tomli  # type: ignore

# Public profiles
PredefinedProfile: TypeAlias = Literal["legacy-additive", "legacy-rebuild", "deep-additive", "deep-rebuild"]

# Private profiles only
_PrivateProfiles: TypeAlias = Literal["no-validation-additive", "no-validation-rebuild"]

# All profiles (union of public and private)
_AllProfiles: TypeAlias = PredefinedProfile | _PrivateProfiles


class ConfigModel(BaseModel):
    model_config = ConfigDict(populate_by_name=True, validate_assignment=True)


class ValidationConfig(ConfigModel):
    """Validation configuration."""

    exclude: list[str] = Field(default_factory=list)
    override: bool = Field(False, description="If enabled, all validators are skipped.")

    def can_run_validator(self, code: str, issue_type: type) -> bool:
        """
        Check if a specific validator should run.

        Args:
            code: Validation code (e.g., "NEAT-DMS-CONTAINER-001")
            issue_type: Issue type (e.g., ModelSyntaxError, ConsistencyError, Recommendation)

        Returns:
            True if validator should run, False otherwise
        """

        is_excluded = self._is_excluded(code, self.exclude)

        if issue_type in [ModelSyntaxError, ConsistencyError] and is_excluded and not self.override:
            print(f"Validator {code} was excluded however it is a critical validator and will still run.")
            return True
        else:
            return not is_excluded

    @classmethod
    def _is_excluded(cls, code: str, patterns: list[str]) -> bool:
        """Check if code matches any pattern (supports wildcards)."""
        for pattern in patterns:
            if "*" in pattern:
                # Split both pattern and code by hyphens
                pattern_parts = pattern.split("-")
                code_parts = code.split("-")

                # Pattern must have same or fewer parts than code
                if len(pattern_parts) > len(code_parts):
                    continue

                # Check if all pattern parts match (allowing wildcards)
                match = True
                for p_part, c_part in zip(pattern_parts, code_parts, strict=False):
                    if p_part != "*" and p_part != c_part:
                        match = False
                        break

                if match:
                    return True
            elif code == pattern:
                return True

        return False

    def __str__(self) -> str:
        """Human-readable configuration summary."""
        if not self.exclude:
            return "All validators enabled"
        return f"Excluded Rules: {', '.join(self.exclude)}"


class ModelingConfig(ConfigModel):
    """Modeling configuration."""

    mode: ModusOperandi = "additive"


class AlphaFlagConfig(ConfigModel):
    """Alpha feature flags configuration."""

    fix_validation_issues: bool = Field(
        default=False,
        description="If enabled, Neat will attempt to automatically fix certain validation issues in the data model.",
    )
    enable_experimental_validators: bool = Field(
        default=False,
        description="If enabled, Neat will run experimental validators that are still in alpha stage.",
    )
    enable_solution_model_creation: bool = Field(
        default=False, description="If enabled, neat.physical_data_model.create() will be available"
    )
    enable_cdf_analysis: bool = Field(default=False, description="If enabled, neat.cdf endpoint will be available.")

    def __setattr__(self, key: str, value: Any) -> None:
        """Set attribute value or raise AttributeError."""
        if key in self.model_fields:
            super().__setattr__(key, value)
        else:
            available_flags = humanize_collection(type(self).model_fields.keys())
            raise UserInputError(f"Alpha flag '{key}' not found. Available flags: {available_flags}.")

    def _repr_html_(self) -> str:
        """HTML representation for Jupyter notebooks."""
        lines = ["<b>Alpha Feature Flags:</b>"]
        for field_name, field in type(self).model_fields.items():
            value = getattr(self, field_name)
            display = "Enabled" if value else "Disabled"
            lines.append(f"<li><b>{field_name}</b>: {display} - {field.description}</li>")
        return "<ul>" + "\n".join(lines) + "</ul>"


class NeatConfig(ConfigModel):
    """NeatSession configuration."""

    profile: str
    validation: ValidationConfig
    modeling: ModelingConfig
    alpha: AlphaFlagConfig = Field(default_factory=AlphaFlagConfig)

    def __str__(self) -> str:
        """Human-readable configuration summary."""
        lines = [
            f"Profile: {self.profile}",
            f"Modeling Mode: {self.modeling.mode}",
            f"Validation: {self.validation}",
        ]
        return "\n".join(lines)

    @classmethod
    def create_predefined(cls, profile: PredefinedProfile = "legacy-additive") -> "NeatConfig":
        """Create NeatConfig from internal profiles.

        Args:
            profile: Profile name to use. Defaults to "legacy-additive".

        !!! note "Predefined Profiles"
            The following predefined profiles are available:

            - `legacy-additive`: Additive modeling with legacy validation rules.
            - `legacy-rebuild`: Rebuild modeling with legacy validation rules.
            - `deep-additive`: Additive modeling with deep validation rules.
            - `deep-rebuild`: Rebuild modeling with deep validation rules.
        """
        available_profiles = internal_profiles()
        if profile not in available_profiles:
            raise UserInputError(
                f"Profile {profile!r} not found. Available predefined profiles: "
                f"{humanize_collection(available_profiles.keys())}."
            )
        return available_profiles[profile]


def internal_profiles() -> dict[_AllProfiles, NeatConfig]:
    """Get internal NeatConfig profile by name."""
    return {
        "legacy-additive": NeatConfig(
            profile="legacy-additive",
            modeling=ModelingConfig(mode="additive"),
            validation=ValidationConfig(
                exclude=[
                    "NEAT-DMS-AI-READINESS-*",
                    "NEAT-DMS-CONNECTIONS-002",
                    "NEAT-DMS-CONNECTIONS-REVERSE-007",
                    "NEAT-DMS-CONNECTIONS-REVERSE-008",
                    "NEAT-DMS-CONSISTENCY-001",
                ]
            ),
        ),
        "legacy-rebuild": NeatConfig(
            profile="legacy-rebuild",
            modeling=ModelingConfig(mode="rebuild"),
            validation=ValidationConfig(
                exclude=[
                    "NEAT-DMS-AI-READINESS-*",
                    "NEAT-DMS-CONNECTIONS-002",
                    "NEAT-DMS-CONNECTIONS-REVERSE-007",
                    "NEAT-DMS-CONNECTIONS-REVERSE-008",
                    "NEAT-DMS-CONSISTENCY-001",
                ]
            ),
        ),
        "deep-additive": NeatConfig(
            profile="deep-additive",
            modeling=ModelingConfig(mode="additive"),
            validation=ValidationConfig(exclude=[]),
        ),
        "deep-rebuild": NeatConfig(
            profile="deep-rebuild",
            modeling=ModelingConfig(mode="rebuild"),
            validation=ValidationConfig(exclude=[]),
        ),
        "no-validation-rebuild": NeatConfig(
            profile="no-validation-rebuild",
            modeling=ModelingConfig(mode="rebuild"),
            validation=ValidationConfig(exclude=["*"], override=True),
        ),
        "no-validation-additive": NeatConfig(
            profile="no-validation-additive",
            modeling=ModelingConfig(mode="additive"),
            validation=ValidationConfig(exclude=["*"], override=True),
        ),
    }


def get_neat_config_from_file(config_file_name: str, profile: str) -> NeatConfig:
    """Get NeatConfig from file or internal profiles.

    Args:
        config_file_name: Path to configuration file.
        profile: Profile name to use.
    Returns:
        NeatConfig instance.
    """

    if not config_file_name.endswith(".toml"):
        raise ValueError("config_file_name must end with '.toml'")

    file_path = Path.cwd() / config_file_name

    if file_path.exists():
        with file_path.open("rb") as f:
            toml = tomli.load(f)

        if "tool" in toml and "neat" in toml["tool"]:
            data = toml["tool"]["neat"]
        elif "neat" in toml:
            data = toml["neat"]
        else:
            raise ValueError("No [tool.neat] or [neat] section found in the configuration file.")

        toml_profile = data.get("profile")
        toml_profiles = data.get("profiles")
        hardcoded_profiles = internal_profiles()

        if toml_profile and toml_profile in hardcoded_profiles:
            raise ValueError(f"Internal profile '{toml_profile}' cannot be used in external configuration file.")

        if toml_profiles and any(p in hardcoded_profiles for p in toml_profiles.keys()):
            raise ValueError(
                "Internal profiles cannot be redefined in external configuration file: "
                f"{set(hardcoded_profiles.keys()).intersection(toml_profiles.keys())}"
            )

        if toml_profile and profile == toml_profile:
            return NeatConfig(**data)
        elif (built_in_profiles := data.get("profiles")) and profile in built_in_profiles:
            return NeatConfig(profile=profile, **data["profiles"][profile])
        else:
            raise ValueError(f"Profile '{profile}' not found in configuration file.")
    else:
        raise FileNotFoundError(f"Configuration file '{file_path}' not found.")
